#

from larch import Model, P, X

def test_utility_function_output():
	k = Model()
	k.utility_ca = P.Aaa * X.Aaa + P.Bbb * X.Bbb + P.Ccc
	k.utility_co[1] = P.Dx1 + P.Dy1 * X.Yyy
	k.utility_co[2] = P.Dx2 + P.Dy2 * X.Yyy

	k.quantity_ca = P.Qaa * X.Aaa + P.Qbb * X.Bbb + P.Qcc

	k.set_values(Aaa=12, Bbb=20, Ccc=2, Dx1=0, Dy1=0.001, Dx2=0.33, Dy2=-0.002)

	import larch.model.linear
	larch.model.linear.use_tooltips = True

	u1 = k.utility_functions(resolve_parameters=False)

	assert u1.tostring() == '<div><table class="floatinghead" style="margin-top:1px;"><thead>' \
							'<tr><th>alt</th><th style="text-align:left;">formula</th></tr></thead>' \
							'<tbody><tr><td>1</td><td style="text-align:left;"><div></div> + ' \
							'<div class="tooltipped">P.Aaa<span class="tooltiptext">12</span></div> * ' \
							'<div class="tooltipped">X.Aaa<span class="tooltiptext">This is Data</span>' \
							'</div><br> + <div class="tooltipped">P.Bbb<span class="tooltiptext">20</span>' \
							'</div> * <div class="tooltipped">X.Bbb<span class="tooltiptext">This is Data' \
							'</span></div><br> + <div class="tooltipped">P.Ccc<span class="tooltiptext">2' \
							'</span></div><br> + <div class="tooltipped">P.Dx1<span class="tooltiptext">0' \
							'</span></div><br> + <div class="tooltipped">P.Dy1<span class="tooltiptext">0.001' \
							'</span></div> * <div class="tooltipped">X.Yyy<span class="tooltiptext">' \
							'This is Data</span></div><br> + log(<br>\xa0\xa0 + <span></span>exp(' \
							'<div class="tooltipped">P.Qaa<span class="tooltiptext">exp(0) = 0</span></div>) ' \
							'* <div class="tooltipped">X.Aaa<span class="tooltiptext">This is Data</span></div>' \
							'<br>\xa0\xa0 + <span></span>exp(<div class="tooltipped">P.Qbb<span class="tooltiptext">' \
							'exp(0) = 0</span></div>) * <div class="tooltipped">X.Bbb<span class="tooltiptext">' \
							'This is Data</span></div><br>\xa0\xa0 + <span></span>exp(<div class="tooltipped">' \
							'P.Qcc<span class="tooltiptext">exp(0) = 0</span></div>)<br>)</td></tr><tr><td>2</td>' \
							'<td style="text-align:left;"><div></div> + <div class="tooltipped">P.Aaa' \
							'<span class="tooltiptext">12</span></div> * <div class="tooltipped">X.Aaa' \
							'<span class="tooltiptext">This is Data</span></div><br> + <div class="tooltipped">' \
							'P.Bbb<span class="tooltiptext">20</span></div> * <div class="tooltipped">X.Bbb' \
							'<span class="tooltiptext">This is Data</span></div><br> + <div class="tooltipped">' \
							'P.Ccc<span class="tooltiptext">2</span></div><br> + <div class="tooltipped">P.Dx2' \
							'<span class="tooltiptext">0.33</span></div><br> + <div class="tooltipped">P.Dy2' \
							'<span class="tooltiptext">-0.002</span></div> * <div class="tooltipped">X.Yyy' \
							'<span class="tooltiptext">This is Data</span></div><br> + log(<br>\xa0\xa0 + ' \
							'<span></span>exp(<div class="tooltipped">P.Qaa<span class="tooltiptext">exp(0) = 0' \
							'</span></div>) * <div class="tooltipped">X.Aaa<span class="tooltiptext">' \
							'This is Data</span></div><br>\xa0\xa0 + <span></span>exp(<div class="tooltipped">' \
							'P.Qbb<span class="tooltiptext">exp(0) = 0</span></div>) * <div class="tooltipped">' \
							'X.Bbb<span class="tooltiptext">This is Data</span></div><br>\xa0\xa0 + <span>' \
							'</span>exp(<div class="tooltipped">P.Qcc<span class="tooltiptext">exp(0) = 0</span>' \
							'</div>)<br>)</td></tr></tbody></table></div>'

	u2 = k.utility_functions(resolve_parameters=True)

	assert u2.tostring() == '<div><table class="floatinghead" style="margin-top:1px;"><thead><tr><th>alt</th>' \
							'<th style="text-align:left;">formula</th></tr></thead><tbody><tr><td>1</td>' \
							'<td style="text-align:left;"><div></div> + <div class="tooltipped">12' \
							'<span class="tooltiptext">P.Aaa</span></div> * <div class="tooltipped">' \
							'X.Aaa<span class="tooltiptext">This is Data</span></div><br> + <div class="tooltipped">' \
							'20<span class="tooltiptext">P.Bbb</span></div> * <div class="tooltipped">X.Bbb' \
							'<span class="tooltiptext">This is Data</span></div><br> + <div class="tooltipped">' \
							'2<span class="tooltiptext">P.Ccc</span></div><br> + <div class="tooltipped">0' \
							'<span class="tooltiptext">P.Dx1</span></div><br> + <div class="tooltipped">0.001' \
							'<span class="tooltiptext">P.Dy1</span></div> * <div class="tooltipped">X.Yyy' \
							'<span class="tooltiptext">This is Data</span></div><br> + log(<br>\xa0\xa0 + <span>' \
							'</span>exp(<div class="tooltipped">0<span class="tooltiptext">exp(P.Qaa)</span></div>) ' \
							'* <div class="tooltipped">X.Aaa<span class="tooltiptext">This is Data</span></div>' \
							'<br>\xa0\xa0 + <span></span>exp(<div class="tooltipped">0<span class="tooltiptext">' \
							'exp(P.Qbb)</span></div>) * <div class="tooltipped">X.Bbb<span class="tooltiptext">' \
							'This is Data</span></div><br>\xa0\xa0 + <span></span>exp(<div class="tooltipped">0' \
							'<span class="tooltiptext">exp(P.Qcc)</span></div>)<br>)</td></tr><tr><td>2</td>' \
							'<td style="text-align:left;"><div></div> + <div class="tooltipped">12' \
							'<span class="tooltiptext">P.Aaa</span></div> * <div class="tooltipped">X.Aaa' \
							'<span class="tooltiptext">This is Data</span></div><br> + <div class="tooltipped">' \
							'20<span class="tooltiptext">P.Bbb</span></div> * <div class="tooltipped">X.Bbb' \
							'<span class="tooltiptext">This is Data</span></div><br> + <div class="tooltipped">2' \
							'<span class="tooltiptext">P.Ccc</span></div><br> + <div class="tooltipped">0.33' \
							'<span class="tooltiptext">P.Dx2</span></div><br> + <div class="tooltipped">-0.002' \
							'<span class="tooltiptext">P.Dy2</span></div> * <div class="tooltipped">X.Yyy' \
							'<span class="tooltiptext">This is Data</span></div><br> + log(<br>\xa0\xa0 + <span>' \
							'</span>exp(<div class="tooltipped">0<span class="tooltiptext">exp(P.Qaa)</span></div>) ' \
							'* <div class="tooltipped">X.Aaa<span class="tooltiptext">This is Data</span></div>' \
							'<br>\xa0\xa0 + <span></span>exp(<div class="tooltipped">0<span class="tooltiptext">' \
							'exp(P.Qbb)</span></div>) * <div class="tooltipped">X.Bbb<span class="tooltiptext">' \
							'This is Data</span></div><br>\xa0\xa0 + <span></span>exp(<div class="tooltipped">' \
							'0<span class="tooltiptext">exp(P.Qcc)</span></div>)<br>)</td></tr></tbody></table></div>'




import numpy, pandas


def define_masks(bitmask_sizes):
	"""

	Parameters
	----------
	bitmask_sizes : iterable

	Returns
	-------
	list
	"""

	shifts = ((numpy.cumsum(bitmask_sizes[::-1]))[::-1]).astype(int)
	shifts[:-1] = shifts[1:]
	shifts[-1] = 0

	masks = [
		((2**bitmask_sizes[b])-1)
		<<shifts[b]
		for b in range(len(bitmask_sizes))
	]

	return masks


def define_min_masks_from_sets( sets ):
	"""

	Parameters
	----------
	sets : iterable of iterable

	Returns
	-------
	list
	"""
	bitmask_sizes = [ int(numpy.ceil(numpy.log2(len(numpy.unique(s))))) for s in sets ]
	return define_masks(bitmask_sizes)


def define_min_masks_from_df( df ):
	"""

	Parameters
	----------
	df : pandas.DataFrame

	Returns
	-------
	list
	"""
	bitmask_sizes = [ int(numpy.ceil(numpy.log2(len(numpy.unique(df[s]))))) for s in df.columns ]
	return define_masks(bitmask_sizes)



import os
import math
import copy
import pandas as pd
import numpy as np
from bs4 import BeautifulSoup

class html_tables(object):

	def __init__(self, raw_html):
		self.url_soup = BeautifulSoup(raw_html, "lxml")

	def read(self):

		self.tables      = []
		self.tables_html = self.url_soup.find_all("table")

		# Parse each table
		for n in range(0, len(self.tables_html)):

			n_cols = 0
			n_rows = 0

			for row in self.tables_html[n].find_all("tr"):
				col_tags = row.find_all(["td", "th"])
				if len(col_tags) > 0:
					n_rows += 1
					if len(col_tags) > n_cols:
						n_cols = len(col_tags)

			# Create dataframe
			df = pd.DataFrame(index = range(0, n_rows), columns = range(0, n_cols))

			# Create list to store rowspan values
			skip_index = [0 for i in range(0, n_cols)]
			this_skip_index = copy.deepcopy(skip_index)

			# Start by iterating over each row in this table...
			row_counter = 0
			for row in self.tables_html[n].find_all("tr"):

				# Skip row if it's blank
				if len(row.find_all(["td", "th"])) == 0:
					pass

				else:

					# Get all cells containing data in this row
					columns = row.find_all(["td", "th"])
					col_dim = []
					row_dim = []
					col_dim_counter = -1
					row_dim_counter = -1
					col_counter = -1
					this_skip_index = copy.deepcopy(skip_index)

					for col in columns:

						# Determine cell dimensions
						colspan = col.get("colspan")
						if colspan is None:
							col_dim.append(1)
						else:
							col_dim.append(int(colspan))
						col_dim_counter += 1

						rowspan = col.get("rowspan")
						if rowspan is None:
							row_dim.append(1)
						else:
							row_dim.append(int(rowspan))
						row_dim_counter += 1

						# Adjust column counter
						if col_counter == -1:
							col_counter = 0
						else:
							col_counter = col_counter + col_dim[col_dim_counter - 1]

						try:
							while skip_index[col_counter] > 0:
								col_counter += 1
						except IndexError:
							from pprint import pprint
							print("~"*50)
							pprint(locals())
							print("~"*50)
							raise
							pass

						# Get cell contents
						cell_data = col.get_text()

						# Insert data into cell
						df.iat[row_counter, col_counter] = cell_data

						# Record column skipping index
						if row_dim[row_dim_counter] > 1:
							this_skip_index[col_counter] = row_dim[row_dim_counter]

				# Adjust row counter
				row_counter += 1

				# Adjust column skipping index
				skip_index = [i - 1 if i > 0 else i for i in this_skip_index]

			# Append dataframe to list of tables
			self.tables.append(df)

		return(self.tables)



def xml_table_to_txt(elem):
	first_table = html_tables(elem.tostring()).read()[0]
	return (first_table.to_string(
		buf=None, columns=None, col_space=None, header=True,
		index=False, na_rep='', formatters=None, float_format=None,
		sparsify=None, index_names=False, justify=None, line_width=None,
		max_rows=None, max_cols=None, show_dimensions=False).partition("\n")[2])





import numpy
import pandas

def tiny_perturbation(arr):
	fr = numpy.frexp(arr)
	perturb = numpy.random.random(fr[0].shape) * 1e-12
	return numpy.ldexp(fr[0]+perturb, fr[1])


def rank_within_cases(df, values, groupdef, output_col=None, tiny_perturb=False):
	if tiny_perturb:
		x = tiny_perturbation(df[values])
	else:
		x = df[values]

	result = x.groupby(df[groupdef]).rank("first", ascending=False)

	if output_col is not None:
		df[output_col] = result

	return result


def probability_to_rank(arr):
	"""Convert a probability array to ranks.

	Parameters
	----------
	arr: array-like
		The probability array to convert.  Each row is a case observation
		and should have probability that sums to 1.

	Returns
	-------
	ranks
		An array or DataFrame (as arr) with ranks.
	"""

	temp = arr.argsort(axis=-1)
	ranks = numpy.empty_like(temp)
	seq = numpy.arange(1,arr.shape[1]+1)[::-1]
	for i in range(arr.shape[0]):
		ranks[i,temp[i]] = seq

	if isinstance(arr, pandas.DataFrame):
		return pandas.DataFrame(ranks, columns=arr.columns, index=arr.index)
	else:
		return ranks


def top_k_accuracy(ranks, choices, k=1):
	"""
	Compute top-N accuracy.

	Parameters
	----------
	ranks: array-like
		Ranked alternatives.  The best alternative in a row should be ranked 1,
		the next 2, and so on.
	choices
		A one-hot encoded or weighted choice array.
	k : int or iterable of ints
		k

	Returns
	-------
	float
		Fraction of rows where the chosen alternative is ranked at k or better.
	"""
	if isinstance(k, (int, numpy.integer)):
		ranks = numpy.asarray(ranks)
		choices = numpy.asarray(choices)
		_1 = (ranks[choices > 0] <= k)
		_2 = (choices[choices > 0])
		return (_1*_2).sum(axis=None) / (choices).sum(axis=None)

	results = {}
	for k_ in k:
		results[k_] = top_k_accuracy(ranks, choices, k=k_)

	return pandas.Series(results)






LIGHT_CONSOLE_LOG_FORMAT = '%(message)s'

from .normal import *

# Remove timestamps from console stream

for entry in logger.handlers:
	if (isinstance(entry, logging.StreamHandler)) and (entry.formatter._fmt == CONSOLE_LOG_FORMAT):
		entry.setFormatter(logging.Formatter(LIGHT_CONSOLE_LOG_FORMAT))





_reserved_names_ = {'_VAULT_', '_BUNCHES_'}

# import numpy
from .pod import Pod
from .podlist import Pods
from .service import DataService
from .h5 import *
from .general import SystematicAlternatives



import warnings
import contextlib

@contextlib.contextmanager
def ignore_warnings():
	with warnings.catch_warnings():
		warnings.simplefilter("ignore")
		yield


class NoDefaultValueWarning(Warning):
	pass

def no_default_value(message):
	#warnings.warn(message, NoDefaultValueWarning, stacklevel=2)
	pass




import numpy
import warnings


class PossibleOverspecification(Warning):
	pass


def compute_possible_overspecification(a, holdfast_vector=None):
	"""

	Parameters
	----------
	a : matrix
		The hessian matrix
	holdfast_vector : vector
		A vectgor of indicators for which row/cols should be ignored as holdfast-ed

	Returns
	-------

	"""
	ret = []
	if holdfast_vector is None:
		holdfast_vector = numpy.zeros(a.shape[0], dtype=bool)
	else:
		holdfast_vector = holdfast_vector.astype(bool, copy=True)
	holdfast_vector |= ((a == 0).all(0))
	a_packed = a[~holdfast_vector, :][:, ~holdfast_vector]
	try:
		eigenvalues_packed, eigenvectors_packed = numpy.linalg.eigh(a_packed)
	except numpy.linalg.linalg.LinAlgError as err:
		return [('LinAlgError', str(err), '')]
	for i in range(len(eigenvalues_packed)):
		if numpy.abs(eigenvalues_packed[i]) < 0.001:
			v = eigenvectors_packed[:, i]
			v = numpy.round(v, 7)
			v_unpacked = numpy.zeros(a.shape[0])
			v_unpacked[~holdfast_vector.astype(bool)] = v
			ret.append((eigenvalues_packed[i], numpy.where(v_unpacked)[0], v_unpacked))
	return ret





from .linear import ParameterRef_C, _what_is
from numbers import Number

_precedence = {
	'lambda': 0,
	'if': 1,
	'or': 2,
	'and': 3,
	'not': 4,
	'compare': 5,
	'|': 6,
	'^': 7,
	'&': 8,
	'>>': 9,
	'+': 10,
	'*': 11,
	'~': 12,
	'**': 13,
}

def _parenthize_repr(obj, local_precedence):
	obj_precedence = getattr(obj, '_precedence', 0)
	if obj_precedence < local_precedence:
		return f"({repr(obj)})"
	else:
		return repr(obj)


class _ParameterOp():

	def set_fmt(self, formatting):
		self._formatting = formatting
		return self

	def value(self, *args):
		raise NotImplementedError('abstract base class')

	def string(self, m):
		"""
		The value of the parameter math from a given model, as a formatted string.

		Parameters
		----------
		m : Model
			The model from which to extract a parameter value.

		Returns
		-------
		str
		"""
		if self._formatting is None:
			return "{:.3g}".format(self.value(m))
		else:
			return self._formatting.format(self.value(m))

	def __add__(self, other):
		if isinstance(other, (ParameterRef_C, _ParameterOp, Number)):
			return ParameterAdd(self, other)
		if hasattr(other, 'as_pmath'):
			try:
				return ParameterAdd(self, other.as_pmath())
			except:
				pass
		raise NotImplementedError(f"{_what_is(self)} + {_what_is(other)}")

	def __radd__(self, other):
		if isinstance(other, (ParameterRef_C, _ParameterOp, Number)):
			return ParameterAdd(other, self)
		if hasattr(other, 'as_pmath'):
			try:
				return ParameterAdd(other.as_pmath(), self)
			except:
				pass
		raise NotImplementedError(f"{_what_is(other)} + {_what_is(self)}")

	def __sub__(self, other):
		if isinstance(other, (ParameterRef_C, _ParameterOp, Number)):
			return ParameterSubtract(self, other)
		if hasattr(other, 'as_pmath'):
			try:
				return ParameterSubtract(self, other.as_pmath())
			except:
				pass
		raise NotImplementedError(f"{_what_is(self)} - {_what_is(other)}")

	def __rsub__(self, other):
		if isinstance(other, (ParameterRef_C, _ParameterOp, Number)):
			return ParameterSubtract(other, self)
		if hasattr(other, 'as_pmath'):
			try:
				return ParameterSubtract(other.as_pmath(), self)
			except:
				pass
		raise NotImplementedError(f"{_what_is(other)} - {_what_is(self)}")

	def __mul__(self, other):
		if isinstance(other, (ParameterRef_C, _ParameterOp, Number)):
			return ParameterMultiply(self, other)
		if hasattr(other, 'as_pmath'):
			try:
				return ParameterMultiply(self, other.as_pmath())
			except:
				pass
		raise NotImplementedError(f"{_what_is(self)} * {_what_is(other)}")

	def __rmul__(self, other):
		if isinstance(other, (ParameterRef_C, _ParameterOp, Number)):
			return ParameterMultiply(other, self)
		if hasattr(other, 'as_pmath'):
			try:
				return ParameterMultiply(other.as_pmath(), self)
			except:
				pass
		raise NotImplementedError(f"{_what_is(other)} * {_what_is(self)}")

	def __truediv__(self, other):
		if isinstance(other, (ParameterRef_C, _ParameterOp, Number)):
			return ParameterDivide(self, other)
		if hasattr(other, 'as_pmath'):
			try:
				return ParameterDivide(self, other.as_pmath())
			except:
				pass
		raise NotImplementedError(f"{_what_is(self)} / {_what_is(other)}")

	def __rtruediv__(self, other):
		if isinstance(other, (ParameterRef_C, _ParameterOp, Number)):
			return ParameterDivide(other, self)
		if hasattr(other, 'as_pmath'):
			try:
				return ParameterDivide(other.as_pmath(), self)
			except:
				pass
		raise NotImplementedError(f"{_what_is(other)} / {_what_is(self)}")


ParameterOp = _ParameterOp

class _ParameterUnaryOp(_ParameterOp):
	_op = '??'
	_precedence = 0

	def __init__(self, operand):
		self._operand = operand
		self._formatting = getattr(operand, '_formatting', None)

	def valid(self, m):
		return self._operand.valid(m)

	def __repr__(self):
		return "{}{}".format(
			self._op,
			_parenthize_repr(self._operand, self._precedence),
		)


class _ParameterBinaryOp(_ParameterOp):
	_op = '??'
	_precedence = 0

	def __init__(self, left, right):
		self._left = left
		self._right = right
		self._formatting = getattr(left, '_formatting', None) or getattr(right, '_formatting', None)

	def valid(self, m):
		if isinstance(self._left, (ParameterRef_C, _ParameterOp)):
			x = self._left.valid(m)
		else:
			x = True
		if isinstance(self._right, (ParameterRef_C, _ParameterOp)):
			x &= self._right.valid(m)
		return x

	def __repr__(self):
		return "{} {} {}".format(
			_parenthize_repr(self._left, self._precedence),
			self._op,
			_parenthize_repr(self._right, self._precedence),
		)

	def __str__(self):
		return "{} {} {}".format(
			_parenthize_repr(self._left, self._precedence),
			self._op,
			_parenthize_repr(self._right, self._precedence),
		)


class ParameterAdd(_ParameterBinaryOp):
	_op = '+'
	_precedence = 10

	def value(self, *args):
		if isinstance(self._left, (ParameterRef_C, _ParameterOp)):
			x = self._left.value(*args)
		else:
			x = self._left
		if isinstance(self._right, (ParameterRef_C, _ParameterOp)):
			x += self._right.value(*args)
		else:
			x += self._right
		return x


class ParameterMultiply(_ParameterBinaryOp):
	_op = '*'
	_precedence = 11

	def value(self, *args):
		if isinstance(self._left, (ParameterRef_C, _ParameterOp)):
			x = self._left.value(*args)
		else:
			x = self._left
		if isinstance(self._right, (ParameterRef_C, _ParameterOp)):
			x *= self._right.value(*args)
		else:
			x *= self._right
		return x


class ParameterSubtract(_ParameterBinaryOp):
	_op = '-'
	_precedence = 10

	def value(self, *args):
		if isinstance(self._left, (ParameterRef_C, _ParameterOp)):
			x = self._left.value(*args)
		else:
			x = self._left
		if isinstance(self._right, (ParameterRef_C, _ParameterOp)):
			x -= self._right.value(*args)
		else:
			x -= self._right
		return x


class ParameterDivide(_ParameterBinaryOp):
	_op = '/'
	_precedence = 11

	def value(self, *args):
		if isinstance(self._left, (ParameterRef_C, _ParameterOp)):
			x = self._left.value(*args)
		else:
			x = self._left
		if isinstance(self._right, (ParameterRef_C, _ParameterOp)):
			x /= self._right.value(*args)
		else:
			x /= self._right
		return x


class ParameterNoop(_ParameterUnaryOp):
	_op = ''
	_precedence = 98

	def value(self, *args):
		return self._operand.value(*args)


class ParameterNegate(_ParameterUnaryOp):
	_op = '-'
	_precedence = 12

	def value(self, *args):
		return -self._operand.value(*args)




from larch.model import Model
import numpy
from pytest import approx
import pandas
import pytest

def qmnl_straw_man_model_1():

	from larch.roles import P,X

	altcodes = (1, 2, 3, 4, 5, 6)
	from larch.data_services.examples import MTC
	dt = MTC()

	p = Model(parameters=[], alts=altcodes, dataservice=dt, graph=None)


	from larch.roles import P,X

	p.utility_ca = (
		+ P('tottime') * X('tottime')
		+ P('totcost') * X('totcost')
	)

	p.utility_co = {
		2: (P('ASC#2') * X('1') + P('hhinc#2') * X('hhinc')),
		3: (P('ASC#3') * X('1') + P('hhinc#3') * X('hhinc')),
		4: (P('ASC#4') * X('1') + P('hhinc#4') * X('hhinc')),
		5: (P('ASC#5') * X('1') + P('hhinc#5') * X('hhinc')),
		6: (P('ASC#6') * X('1') + P('hhinc#6') * X('hhinc')),
	}

	p.quantity_ca = (
		+ P("FakeSizeAlt") * X('altnum+1')
		+ P("FakeSizeIvtt") * X('ivtt+1')
	)

	p.availability_var = '_avail_'
	p.choice_ca_var = '_choice_'

	p.load_data()
	return p


def test_qmnl():
	pq = qmnl_straw_man_model_1()
	assert( pq.loglike() == approx(-8486.55377320886))

	dll = pq.d_loglike()

	names = ['ASC#2', 'ASC#5', 'hhinc#3', 'ASC#4', 'ASC#6', 'hhinc#6', 'ASC#3',
       'hhinc#4', 'tottime', 'hhinc#2', 'hhinc#5', 'totcost', 'FakeSizeIvtt',
       'FakeSizeAlt']

	correct = pandas.Series(
		[-6.76598075e+02, -4.43123432e+02, -6.73124640e+04,
		 -4.91818432e+02, 3.97095316e+00, -1.38966274e+03,
		 -1.16626503e+03, -3.06932152e+04, -4.87328610e+04,
		 -4.02490548e+04, -2.72367637e+04, 1.45788603e+05,
		 8.69414603e+01, -8.69414603e+01],
			index = names)

	for n in names:
		assert (correct.loc[n] == approx(dll.loc[n]))

@pytest.mark.skip(reason="test is for manual use to check multithreading")
def test_saturate_cpu(howlong=15):
	pq = qmnl_straw_man_model_1()
	pq.initialize_graph(alternative_codes=[1,2,3,4,5,6])
	pq.graph.add_node(10, children=(5, 6), parameter='MU_nonmotor')
	pq.graph.add_node(11, children=(1, 2, 3), parameter='MU_car')
	assert( pq.loglike() == approx(-8486.55377320886))

	import time
	starttime = time.time()
	while time.time() < starttime + howlong:
		dll = pq.d_loglike()



#
import numpy
import pandas
import larch
from larch import Model, P, X
import larch.exampville
import os
from pytest import approx
from .stored_dataframes import stable_df

def test_ch_av_summary_output():

	skims = larch.OMX(larch.exampville.files.skims, mode='r')
	hh = pandas.read_csv(larch.exampville.files.hh)
	pp = pandas.read_csv(larch.exampville.files.person)
	tour = pandas.read_csv(larch.exampville.files.tour)

	pp_col = ['PERSONID', 'HHID', 'HHIDX', 'AGE', 'WORKS',
       'N_WORK_TOURS', 'N_OTHER_TOURS', 'N_TOURS', 'N_TRIPS', 'N_TRIPS_HBW',
       'N_TRIPS_HBO', 'N_TRIPS_NHB']

	raw = tour.merge(hh, on='HHID').merge(pp[pp_col], on=('HHID', 'PERSONID'))
	raw["HOMETAZi"] = raw["HOMETAZ"] - 1
	raw["DTAZi"] = raw["DTAZ"] - 1

	raw = raw[raw.TOURPURP == 1]

	f_tour = raw.join(
		skims.get_rc_dataframe(
			raw.HOMETAZi, raw.DTAZi,
		)
	)

	DA = 1
	SR = 2
	Walk = 3
	Bike = 4
	Transit = 5

	dfs = larch.DataFrames(
		co=f_tour,
		alt_codes=[DA, SR, Walk, Bike, Transit],
		alt_names=['DA', 'SR', 'Walk', 'Bike', 'Transit'],
	)

	m = larch.Model(dataservice=dfs)
	m.title = "Exampville Work Tour Mode Choice v1"

	m.utility_co[DA] = (
			+ P.InVehTime * X.AUTO_TIME
			+ P.Cost * X.AUTO_COST  # dollars per mile
	)

	m.utility_co[SR] = (
			+ P.ASC_SR
			+ P.InVehTime * X.AUTO_TIME
			+ P.Cost * (X.AUTO_COST * 0.5)  # dollars per mile, half share
			+ P("HighInc:SR") * X("INCOME>75000")
	)

	m.utility_co[Walk] = (
			+ P.ASC_Walk
			+ P.NonMotorTime * X.WALK_TIME
			+ P("HighInc:Walk") * X("INCOME>75000")
	)

	m.utility_co[Bike] = (
			+ P.ASC_Bike
			+ P.NonMotorTime * X.BIKE_TIME
			+ P("HighInc:Bike") * X("INCOME>75000")
	)

	m.utility_co[Transit] = (
			+ P.ASC_Transit
			+ P.InVehTime * X.TRANSIT_IVTT
			+ P.OutVehTime * X.TRANSIT_OVTT
			+ P.Cost * X.TRANSIT_FARE
			+ P("HighInc:Transit") * X("INCOME>75000")
	)

	# No choice or avail data set
	m.load_data()
	q = m.dataframes.choice_avail_summary()
	assert numpy.array_equal(q.columns, ['name', 'chosen', 'available'])
	assert q.index.identical(pandas.Index([1, 2, 3, 4, 5, '< Total All Alternatives >'], dtype='object'))
	assert numpy.array_equal(q.values, [
		['DA', None, None],
		['SR', None, None],
		['Walk', None, None],
		['Bike', None, None],
		['Transit', None, None],
		['', 0, ''],
	])

	# Reasonable choice and avail data set
	m.choice_co_code = 'TOURMODE'
	m.availability_co_vars = {
		DA: 'AGE >= 16',
		SR: '1',
		Walk: 'WALK_TIME < 60',
		Bike: 'BIKE_TIME < 60',
		Transit: 'TRANSIT_FARE>0',
	}
	m.load_data()
	q = m.dataframes.choice_avail_summary()
	assert numpy.array_equal(q.columns, ['name', 'chosen', 'available'])
	assert q.index.identical(pandas.Index([1, 2, 3, 4, 5, '< Total All Alternatives >'], dtype='object'))
	assert numpy.array_equal(q['name'].values, ['DA', 'SR', 'Walk', 'Bike', 'Transit', ''])
	assert numpy.array_equal(q['chosen'].values, [6052.,  810.,  196.,   72.,  434., 7564.])
	assert numpy.array_equal(q['available'].values,
							 numpy.array([7564.0, 7564.0, 4179.0, 7564.0, 4199.0, ''], dtype=object))

	# Unreasonable choice and avail data set
	m.choice_co_code = 'TOURMODE'
	m.availability_co_vars = {
		DA: 'AGE >= 26',
		SR: '1',
		Walk: 'WALK_TIME < 60',
		Bike: 'BIKE_TIME < 60',
		Transit: 'TRANSIT_FARE>0',
	}
	m.load_data()
	q = m.dataframes.choice_avail_summary()
	assert numpy.array_equal(q.columns, ['name', 'chosen', 'available', 'chosen but not available'])
	assert q.index.identical(pandas.Index([1, 2, 3, 4, 5, '< Total All Alternatives >'], dtype='object'))
	assert numpy.array_equal(q['name'].values, ['DA', 'SR', 'Walk', 'Bike', 'Transit', ''])
	assert numpy.array_equal(q['chosen'].values, [6052.,  810.,  196.,   72.,  434., 7564.])
	assert numpy.array_equal(q['available'].values,
							 numpy.array([6376.0, 7564.0, 4179.0, 7564.0, 4199.0, ''], dtype=object))
	assert numpy.array_equal(q['chosen but not available'].values, [942.0, 0.0, 0.0, 0.0, 0.0, 942.0])



def test_excel_metadata():

	import larch.util.excel
	if larch.util.excel.xlsxwriter is not None:
		from larch.util.excel import _make_excel_writer

		from larch.util.temporaryfile import TemporaryDirectory

		tempdir = TemporaryDirectory()
		os.path.join(tempdir, 'larchtest.xlsx')

		m = larch.example(1)
		m.load_data()
		m.loglike_null()
		m.maximize_loglike()

		xl = _make_excel_writer(m, os.path.join(tempdir, 'larchtest.xlsx'), save_now=False)
		# xl.add_metadata('self', m) # now automatic
		xl.add_metadata('short', 123)
		xl.add_metadata('plain', 'text')
		xl.save()

		md = larch.util.excel.load_metadata(os.path.join(tempdir, 'larchtest.xlsx'))
		assert len(md) == 3
		assert isinstance(md['_self_'], larch.Model)
		assert md['short'] == 123
		assert md['plain'] == 'text'

		md_m = larch.util.excel.load_metadata(os.path.join(tempdir, 'larchtest.xlsx'), '_self_')
		assert isinstance(md_m, larch.Model)

		assert larch.util.excel.load_metadata(os.path.join(tempdir, 'larchtest.xlsx'), 'short') == 123

		from numpy import inf
		assert md_m._get_cached_loglike_values() == approx(
			{'nil': 0.0, 'null': -7309.600971749679, 'constants_only': 0.0, 'best': -inf})

		md_m.estimation_statistics()


def test_parameter_summary():

	m = larch.example(1)
	m.load_data()
	m.loglike_null()
	m.set_values(**{
		'ASC_BIKE': -2.3763275319243244,
		'ASC_SR2': -2.1780143286612037,
		'ASC_SR3P': -3.725078388760564,
		'ASC_TRAN': -0.6708609582690096,
		'ASC_WALK': -0.20677521181801753,
		'hhinc#2': -0.0021699381002406883,
		'hhinc#3': 0.0003577067151217295,
		'hhinc#4': -0.00528632366072714,
		'hhinc#5': -0.012807975284603574,
		'hhinc#6': -0.009686302933787567,
		'totcost': -0.00492023540098787,
		'tottime': -0.05134209452571549,
	})
	m.loglike()
	assert m.parameter_summary('xml').tostring() == (
		'<div><table><thead><tr><th style="text-align: left;">Category</th><th style="text-align: left;">Parameter</th>'
		'<th>Value</th><th>Null Value</th></tr></thead><tbody><tr><th rowspan="2" style="vertical-align: top; text-alig'
		'n: left;">LOS</th><th style="vertical-align: top; text-align: left;">totcost</th><td>-0.00492</td><td>0.0</td>'
		'</tr><tr><th style="vertical-align: top; text-align: left;">tottime</th><td>-0.05134</td><td>0.0</td></tr><tr>'
		'<th rowspan="5" style="vertical-align: top; text-align: left;">ASCs</th><th style="vertical-align: top; text-a'
		'lign: left;">ASC_BIKE</th><td>-2.376</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: left'
		';">ASC_SR2</th><td>-2.178</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: left;">ASC_SR3P'
		'</th><td>-3.725</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: left;">ASC_TRAN</th><td>-'
		'0.6709</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: left;">ASC_WALK</th><td>-0.2068</t'
		'd><td>0.0</td></tr><tr><th rowspan="5" style="vertical-align: top; text-align: left;">Income</th><th style="ve'
		'rtical-align: top; text-align: left;">hhinc#2</th><td>-0.00217</td><td>0.0</td></tr><tr><th style="vertical-al'
		'ign: top; text-align: left;">hhinc#3</th><td>0.0003577</td><td>0.0</td></tr><tr><th style="vertical-align: top'
		'; text-align: left;">hhinc#4</th><td>-0.005286</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-a'
		'lign: left;">hhinc#5</th><td>-0.01281</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: lef'
		't;">hhinc#6</th><td>-0.009686</td><td>0.0</td></tr></tbody></table></div>')
	m.ordering = ()
	assert m.parameter_summary('xml').tostring() == (
		'<div><table><thead><tr><th style="text-align: left;">Parameter</th><th>Value</th><th>Null Value</th></tr></the'
		'ad><tbody><tr><th style="vertical-align: top; text-align: left;">ASC_BIKE</th><td>-2.376</td><td>0.0</td></tr>'
		'<tr><th style="vertical-align: top; text-align: left;">ASC_SR2</th><td>-2.178</td><td>0.0</td></tr><tr><th sty'
		'le="vertical-align: top; text-align: left;">ASC_SR3P</th><td>-3.725</td><td>0.0</td></tr><tr><th style="vertic'
		'al-align: top; text-align: left;">ASC_TRAN</th><td>-0.6709</td><td>0.0</td></tr><tr><th style="vertical-align:'
		' top; text-align: left;">ASC_WALK</th><td>-0.2068</td><td>0.0</td></tr><tr><th style="vertical-align: top; tex'
		't-align: left;">hhinc#2</th><td>-0.00217</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: '
		'left;">hhinc#3</th><td>0.0003577</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: left;">h'
		'hinc#4</th><td>-0.005286</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: left;">hhinc#5</'
		'th><td>-0.01281</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: left;">hhinc#6</th><td>-0'
		'.009686</td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: left;">totcost</th><td>-0.00492</'
		'td><td>0.0</td></tr><tr><th style="vertical-align: top; text-align: left;">tottime</th><td>-0.05134</td><td>0.'
		'0</td></tr></tbody></table></div>')


def test_joint_parameter_summary():
	m = larch.example(1)
	m.load_data()
	m.loglike_null()
	m.set_values(**{
		'ASC_BIKE': -2.3763275319243244,
		'ASC_SR2': -2.1780143286612037,
		'ASC_SR3P': -3.725078388760564,
		'ASC_TRAN': -0.6708609582690096,
		'ASC_WALK': -0.20677521181801753,
		'hhinc#2': -0.0021699381002406883,
		'hhinc#3': 0.0003577067151217295,
		'hhinc#4': -0.00528632366072714,
		'hhinc#5': -0.012807975284603574,
		'hhinc#6': -0.009686302933787567,
		'totcost': -0.00492023540098787,
		'tottime': -0.05134209452571549,
	})
	m.loglike()
	m.maximize_loglike()

	m_c = larch.Model(dataservice=m.dataservice)
	m_c.choice_ca_var = '_choice_'
	m_c.availability_var = '_avail_'
	m_c.utility_co[2] = P("ASC_SR2")
	m_c.utility_co[3] = P("ASC_SR3P")
	m_c.utility_co[4] = P("ASC_TRAN")
	m_c.utility_co[5] = P("ASC_BIKE")
	m_c.utility_co[6] = P("ASC_WALK")
	m_c.title = "Constants Only"
	m_c.load_data()
	m_c.loglike_null()
	m_c.maximize_loglike()

	from larch.util.summary import joint_parameter_summary
	stable_df(joint_parameter_summary([m, m_c], bases=[m_c]), 'joint_parameter_summary')



import logging
import sys
from contextlib import contextmanager
import time
from ..util.timesize import timesize_stack

FILE_LOG_FORMAT = '%(name)s.%(levelname)s: %(message)s'
CONSOLE_LOG_FORMAT = '[%(asctime)s] %(name)s.%(levelname)s: %(message)s'

DEFAULT_LOG_LEVEL = logging.WARNING

from . import logger_name

def log_to_console(level=None):

	if level is None:
		level = DEFAULT_LOG_LEVEL

	logger = logging.getLogger(logger_name)

	# avoid creation of multiple stream handlers for logging to console
	for entry in logger.handlers:
		if (isinstance(entry, logging.StreamHandler)) and (entry.formatter._fmt == CONSOLE_LOG_FORMAT):
			return logger

	console_handler = logging.StreamHandler(stream=sys.stdout)
	console_handler.setLevel(level)
	console_handler.setFormatter(logging.Formatter(CONSOLE_LOG_FORMAT))
	logger.addHandler(console_handler)
	if level < logger.getEffectiveLevel():
		logger.setLevel(level)

	return logger


def log_to_file(filename, level=None):

	if level is None:
		level = DEFAULT_LOG_LEVEL

	logger = logging.getLogger(logger_name)

	# avoid creation of multiple file handlers for logging to the same file
	for entry in logger.handlers:
		if (isinstance(entry, logging.FileHandler)) and (entry.baseFilename == filename):
			return logger

	file_handler = logging.FileHandler(filename)
	file_handler.setLevel(level)
	file_handler.setFormatter(logging.Formatter(FILE_LOG_FORMAT))
	logger.addHandler(file_handler)

	return logger


logger = log = log_to_console()


@contextmanager
def timing_log(label=''):
	start_time = time.time()
	log.critical(f"<TIME BEGINS> {label}")
	try:
		yield
	except:
		log.critical(f"<TIME ERROR!> {label} <{timesize_stack(time.time()-start_time)}>")
		raise
	else:
		log.critical(f"< TIME ENDS > {label} <{timesize_stack(time.time()-start_time)}>")

class TimingLog:

	def __init__(self, label='', log=None, level=50):
		global logger
		if log is None:
			log = logger
		self.label = label
		self.log = log
		self.level = level
		self.split_time = None
		self.current_task = ''

	def __enter__(self):
		self.start_time = time.time()
		self.log.log(self.level, f"<BEGIN> {self.label}")
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		now = time.time()
		if self.split_time is not None:
			self.log.log(self.level, f"<SPLIT> {self.label} / Final <{timesize_stack(now - self.split_time)}>")
		if exc_type is None:
			self.log.log(self.level, f"<-END-> {self.label} <{timesize_stack(now - self.start_time)}>")
		else:
			self.log.log(self.level, f"<ERROR> {self.label} <{timesize_stack(now - self.start_time)}>")

	def split(self, note=''):
		if self.split_time is None:
			self.split_time = self.start_time
		now = time.time()
		if note:
			note = " / " + note
		self.log.log(self.level, f"<SPLIT> {self.label}{note} <{timesize_stack(now - self.split_time)}>")
		self.split_time = now





import numpy as np
import pandas as pd
import re
import itertools
from larch import P, X, DataFrames, Model
from larch.model.model_group import ModelGroup
from larch.util import Dict

import logging
from larch.log import logger_name

_logger = logging.getLogger(logger_name)


def generate_alternatives(n_persons):
    """
    Generate a dictionary of CDAP alternatives.

    The keys are the names of the patterns, and
    the values are the alternative code numbers.

    Parameters
    ----------
    n_persons : int

    Returns
    -------
    Dict
    """
    basic_patterns = ["M", "N", "H"]
    alt_names = list(
        "".join(i) for i in itertools.product(basic_patterns, repeat=n_persons)
    )
    alt_codes = np.arange(1, len(alt_names) + 1)
    return Dict(zip(alt_names, alt_codes))


def apply_replacements(expression, prefix, tokens):
    """
    Convert general person terms to specific person terms for the CDAP model.

    Parameters
    ----------
    expression : str
        An expression from the "Expression" column
        of cdap_INDIV_AND_HHSIZE1_SPEC.csv, or similar.
    prefix : str
        A prefix to attach to each token in `expression`.
    tokens : list-like of str
        A list of tokens to edit within an expression.

    Returns
    -------
    expression : str
        The modified expression
    """
    for i in tokens:
        expression = re.sub(fr"\b{i}\b", f"{prefix}_{i}", expression)
    return expression


def cdap_base_utility_by_person(model, n_persons, spec, alts=None, value_tokens=()):
    """
    Build the base utility by person for each pattern.

    Parameters
    ----------
    model : larch.Model
    n_persons : int
    spec : pandas.DataFrame
        The base utility by person spec provided by
        the ActivitySim framework.
    alts : dict, optional
        The keys are the names of the patterns, and
        the values are the alternative code numbers,
        as created by `generate_alternatives`.  If not
        given, the alts are automatically regenerated
        using that function.
    value_tokens : list-like of str, optional
        A list of tokens to edit within an the expressions,
        generally the column names of the provided values
        from the estimation data bundle.  Only used when
        `n_persons` is more than 1.
    """
    if n_persons == 1:
        for i in spec.index:
            if not pd.isna(spec.loc[i, "M"]):
                model.utility_co[1] += X(spec.Expression[i]) * P(spec.loc[i, "M"])
            if not pd.isna(spec.loc[i, "N"]):
                model.utility_co[2] += X(spec.Expression[i]) * P(spec.loc[i, "N"])
            if not pd.isna(spec.loc[i, "H"]):
                model.utility_co[3] += X(spec.Expression[i]) * P(spec.loc[i, "H"])
    else:
        if alts is None:
            alts = generate_alternatives(n_persons)
        person_numbers = range(1, n_persons + 1)
        for pnum in person_numbers:
            for i in spec.index:
                for aname, anum in alts.items():
                    z = pnum - 1
                    if not pd.isna(spec.loc[i, aname[z]]):
                        x = apply_replacements(
                            spec.Expression[i], f"p{pnum}", value_tokens
                        )
                        model.utility_co[anum] += X(x) * P(spec.loc[i, aname[z]])


def interact_pattern(n_persons, select_persons, tag):
    """
    Compile a regex pattern to match CDAP alternatives.

    Parameters
    ----------
    n_persons : int
    select_persons : list-like of int
        The persons to be selected.
    tag : str
        The activity letter, currently one of {M,N,H}.

    Returns
    -------
    re.compile
    """

    pattern = ""
    p = 1
    while len(pattern) < n_persons:
        pattern += tag if p in select_persons else "."
        p += 1
    return re.compile(pattern)


def cdap_interaction_utility(model, n_persons, alts, interaction_coef):

    person_numbers = list(range(1, n_persons + 1))

    for (cardinality, activity), coefs in interaction_coef.groupby(
        ["cardinality", "activity"]
    ):
        _logger.info(f"{n_persons} person households, interaction cardinality {cardinality}, activity {activity}")
        if cardinality > n_persons:
            continue
        elif cardinality == n_persons:
            this_aname = activity * n_persons
            this_altnum = alts[this_aname]
            for rowindex, row in coefs.iterrows():
                expression = "&".join(
                    f"(p{p}_ptype == {t})"
                    for (p, t) in zip(person_numbers, row.interaction_ptypes)
                    if t != "*"
                )
                if expression:
                    linear_component = X(expression) * P(row.coefficient)
                else:
                    linear_component = P(row.coefficient)
                _logger.debug(
                    f"utility_co[{this_altnum} {this_aname}] += {linear_component}"
                )
                model.utility_co[this_altnum] += linear_component
        elif cardinality < n_persons:
            for combo in itertools.combinations(person_numbers, cardinality):
                pattern = interact_pattern(n_persons, combo, activity)
                for aname, anum in alts.items():
                    if pattern.match(aname):
                        for rowindex, row in coefs.iterrows():
                            expression = "&".join(
                                f"(p{p}_ptype == {t})"
                                for (p, t) in zip(combo, row.interaction_ptypes)
                                if t != "*"
                            )
                            # interaction terms without ptypes (i.e. with wildcards)
                            # only apply when the household size matches the cardinality
                            if expression != "":
                                linear_component = X(expression) * P(row.coefficient)
                                _logger.debug(
                                    f"utility_co[{anum} {aname}] += {linear_component}"
                                )
                                model.utility_co[anum] += linear_component


def cdap_split_data(households, values):
    if "cdap_rank" not in values:
        raise ValueError("assign cdap_rank to values first")
    # only process the first 5 household members
    values = values[values.cdap_rank <= 5]
    cdap_data = {}
    for hhsize, hhs_part in households.groupby(households.hhsize.clip(1, 5)):
        if hhsize == 1:
            v = pd.merge(values, hhs_part.household_id, on="household_id").set_index(
                "household_id"
            )
        else:
            v = (
                pd.merge(values, hhs_part.household_id, on="household_id")
                .set_index(["household_id", "cdap_rank"])
                .unstack()
            )
            v.columns = [f"p{i[1]}_{i[0]}" for i in v.columns]
            for agglom in ["override_choice", "model_choice"]:
                v[agglom] = v[[f"p{p}_{agglom}" for p in range(1, hhsize + 1)]].sum(1)
        cdap_data[hhsize] = v
    return cdap_data


def cdap_dataframes(households, values):
    data = cdap_split_data(households, values)
    dfs = {}
    for hhsize in data.keys():
        alts = generate_alternatives(hhsize)
        dfs[hhsize] = DataFrames(
            co=data[hhsize],
            alt_names=alts.keys(),
            alt_codes=alts.values(),
            av=1,
            ch=data[hhsize].override_choice.map(alts),
        )
    return dfs


def cdap_model(households, values, spec1, interaction_coef, coefficients):
    cdap_data = cdap_dataframes(households, values)
    m = {}
    _logger.info(f"building for model 1")
    m[1] = Model(dataservice=cdap_data[1])
    cdap_base_utility_by_person(m[1], n_persons=1, spec=spec1)
    m[1].choice_any = True
    m[1].availability_any = True

    # Add cardinality into interaction_coef if not present
    if 'cardinality' not in interaction_coef:
        interaction_coef['cardinality'] = interaction_coef['interaction_ptypes'].str.len()
    for s in [2, 3, 4, 5]:
        _logger.info(f"building for model {s}")
        m[s] = Model(dataservice=cdap_data[s])
        alts = generate_alternatives(s)
        cdap_base_utility_by_person(m[s], s, spec1, alts, values.columns)
        cdap_interaction_utility(m[s], s, alts, interaction_coef)
        m[s].choice_any = True
        m[s].availability_any = True

    result = ModelGroup(m.values())
    from .general import apply_coefficients
    apply_coefficients(coefficients, result)
    return result




import os
import glob

warehouse_directory = os.path.dirname(__file__)

def file_list():
	return list(glob.glob(os.path.join(warehouse_directory,"*")))

def example_file(f, missing_ok=False, rel=False):
	"""
	Get the file path of a particular example file.

	All example files are stored in larch/data_warehouse.

	Parameters
	----------
	f : str
		The base filename to find.
	missing_ok : bool, default False
		Do not raise a FileNotFoundError if the file is missing. This
		is generally only useful in larch scripts that will create the
		example file, not for end users, who generally should not
		create new data files in the data_warehouse directory.
	rel : bool or str, default False
		Return the file path relative to this directory, or if given as
		True, relative to the current working directory.

	Returns
	-------
	str
	"""
	fi = os.path.join(warehouse_directory,f)
	if not os.path.exists(fi) and not missing_ok:
		candidates = list(glob.glob(os.path.join(warehouse_directory,f"*{f}*")))
		if len(candidates)==1:
			return candidates[0]
		raise FileNotFoundError(f)
	if rel is True:
		rel = os.getcwd()
	if rel:
		fi = os.path.relpath(fi, start=rel)
	return fi







# This file is under the Creative Commons Attribution Share Alike license
# https://creativecommons.org/licenses/by-sa/3.0/
# original source:
# https://stackoverflow.com/questions/20643184/using-python-threads-to-make-thousands-of-calls-to-a-slow-api-with-a-rate-limit

from collections.abc import Iterator
from threading import Lock
import time
import functools


class BlockingRateLimiter(Iterator):
	"""Iterator that yields a value at most once every 'interval' seconds."""
	def __init__(self, interval):
		self.lock = Lock()
		self.interval = interval
		self.next_yield = 0

	def __next__(self):
		with self.lock:
			t = time.monotonic()
			if t < self.next_yield:
				time.sleep(self.next_yield - t)
				t = time.monotonic()
			self.next_yield = t + self.interval


class NonBlockingRateLimiter():
	def __init__(self, interval):
		self.lock = Lock()
		self.interval = interval
		self.next_greenlight = 0

	def __call__(self, fn):
		@functools.wraps(fn)
		def decorated(*args, **kwargs):
			t = time.monotonic()
			if t >= self.next_greenlight:
				with self.lock:
					self.next_greenlight = t + self.interval
					fn(*args, **kwargs)
		return decorated

	def __bool__(self):
		t = time.monotonic()
		if t >= self.next_greenlight:
			with self.lock:
				self.next_greenlight = t + self.interval
				return True
		return False


_global_rate_limiters = {}

def GlobalRateLimiter(tag, interval=1, wait_now=True):
	global _global_rate_limiters
	if tag not in _global_rate_limiters:
		_global_rate_limiters[tag] = RateLimiter(interval)
	if wait_now:
		return next(_global_rate_limiters[tag])




import pandas, numpy
from ..data_services.h5.h5pod.generic import CArray

def tally(arr):

	if isinstance(arr, CArray):
		arr_ = arr
		arr = arr[:]
		values, freqs = numpy.unique(arr[~pandas.isnull(arr)], return_counts=True)
		try:
			d = arr_.DICTIONARY
		except:
			d = {}
		s = pandas.Series(freqs, index=[d.get(i,i) for i in values]).sort_values(ascending=False)
		nan_count = pandas.isnull(arr).sum()
		if nan_count:
			s[numpy.NaN] = nan_count
		return s
	else:
		values, freqs = numpy.unique(  arr[~pandas.isnull(arr)], return_counts=True)
		s = pandas.Series(freqs, index=values).sort_values(ascending=False)
		nan_count = pandas.isnull(arr).sum()
		if nan_count:
			s[numpy.NaN] = nan_count
		return s

def tally_weighted(arr, wgt):

	if isinstance(arr, CArray):
		raise NotImplementedError()
	else:
		arr = arr.reshape(-1)
		wgt = wgt.reshape(-1)

		values, inv = numpy.unique(arr[~pandas.isnull(arr)], return_inverse=True)
		s = pandas.Series(numpy.bincount(inv, wgt), index=values).sort_values(ascending=False)
		nan_count = wgt[pandas.isnull(arr)].sum()
		if nan_count:
			s[numpy.NaN] = nan_count
		return s


from .signal_dict import SignalDict

import pprint
import datetime
from .addict_yaml import Dict
from .data_expansion import piece


def _prettyprint_fallback(x):
	if isinstance(x, dictx):
		return repr(x)
	else:
		return pprint.pformat(x)



class dictx(dict):
	"""Python dict with attribute access and xml output."""

	def __repr__(self):
		if self.keys():
			m = max(map(len, list(str(_) for _ in self.keys()))) + 1
			return '\n'.join(['┣'+str(k).rjust(m) + ': ' + _prettyprint_fallback(v).replace('\n','\n┃'+' '*(m+2)) for k, v in self.items()])
		else:
			return self.__class__.__name__ + "()"

	def __getattr__(self, item):
		if item[0]=="_" and item[-1]=="_":
			raise AttributeError(item)
		return self.__getitem__(item)

	def __setattr__(self, name, value):
		if hasattr(dictx, name):
			raise AttributeError("'dictx' object attribute "
								 "'{0}' is read-only".format(name))
		else:
			self[name] = value

	def __xml__(self):
		from xmle import Elem, Show
		x = Elem('div')
		t = x.elem('table', style="margin-top:1px;")
		if len(self):
			tr = t.elem('tr')
			tr.elem('th', text="key")
			tr.elem('th', text='value', style='text-align:left;')
			for k,v in self.items():
				tr = t.elem('tr')
				tr.elem('td', text=str(k))
				try:
					if isinstance(v, str) and '\n' not in v and v[:1] != "#":
						raise ValueError()
					v_ = Show(v)
				except (AttributeError, ValueError):
					plaintext = pprint.pformat(v)
					if isinstance(v, datetime.timedelta):
						plaintext = str(v)
					if "\n" in plaintext:
						tr.elem('td', style='text-align:left;').elem('pre', text=plaintext)
					else:
						tr.elem('td', style='text-align:left;', text=plaintext)
				else:
					tr.elem('td', style='text-align:left;') << v_
		else:
			tr = t.elem('tr')
			tr.elem('td', text="<empty>")
		return x

	def _repr_html_(self):
		return self.__xml__().tostring()


dicta = Dict


# Add statistics to pandas.DataFrame and pandas.Series
import pandas
from .statistics import statistics_for_dataframe, statistics_for_array5, uniques, invmap
pandas.DataFrame.statistics = statistics_for_dataframe
pandas.Series.statistics = statistics_for_array5
pandas.Series.uniques = uniques
pandas.Series.invmap = invmap

from .dataframe import compute
pandas.DataFrame.compute = compute




import apsw
import numpy
import pandas

from .sqlite_arrays import _sqlite_array_2d_float64

class Connection(apsw.Connection):

	def read_data(self, query, dtype=numpy.float64, index_col=None):
		if dtype==numpy.float64:
			return _sqlite_array_2d_float64(self.sqlite3pointer(), query, index_col=index_col)
		raise TypeError("cannot read in {}".format(dtype))






import numpy
import pandas
from typing import MutableMapping

from ..util import Dict
from ..dataframes import DataFrames, MissingDataError
from ..exceptions import ParameterNotInModelWarning
from .linear import ParameterRef_C
from ..general_precision import l4_float_dtype
from ..model import persist_flags
from ..model.abstract_model import AbstractChoiceModel

def sync_frames(*models):
	"""
	Synchronize model parameter frames.

	Parameters
	----------
	*models : Sequence[Model]
	"""
	# check if all frames are already in sync
	in_sync = True
	pf1 = models[0]._frame
	for m in models[1:]:
		if m._frame is not pf1:
			in_sync = False
	if not in_sync:
		joined = pandas.concat([m._frame for m in models], sort=False)
		joined = joined[~joined.index.duplicated(keep='first')]
		for m in models:
			m.set_frame(joined)

class LatentClassModel(AbstractChoiceModel):
	"""
	A latent class model.

	Parameters
	----------
	k_membership : Model
		The class membership model.
	k_models : MutableMapping[int,Model]
		The individual class choice models.
	dataservice : DataService
		The dataservice used to populate data for this model.
	title : str, optional
		A descriptive title for use in reporting.
	frame : pandas.DataFrame, optional
		Initialize the parameterframe with this.
	"""

	def __init__(
			self,
			k_membership,
			k_models,
			*,
			dataservice=None,
			title=None,
			frame=None,
			constraints=None,
	):
		assert isinstance(k_membership, AbstractChoiceModel)
		if len(getattr(k_membership, 'utility_ca', [])):
			raise ValueError("the class membership model cannot include `utility_ca`, only `utility_co`")

		self._k_membership = k_membership
		self._k_membership._model_does_not_require_choice = True
		if not isinstance(k_models, MutableMapping):
			raise ValueError(f'k_models must be a MutableMapping, not {type(k_models)}')
		self._k_models = k_models
		self._dataservice = dataservice
		if self._dataservice is None:
			self._dataservice = k_membership.dataservice
		else:
			if not(k_membership.dataservice is None or k_membership.dataservice is self._dataservice):
				raise ValueError("dataservice for all constituent models must be the same")
		for m in self._k_models.values():
			if self._dataservice is None:
				self._dataservice = m.dataservice
			if not (m.dataservice is None or m.dataservice is self._dataservice):
				raise ValueError("dataservice for all constituent models must be the same")
		self._dataframes = None
		self._mangled = True
		self.constraints = constraints
		super().__init__(
			parameters=None,
			frame=frame,
			title=title,
		)
		self.unmangle()

	def _k_model_names(self):
		return list(sorted(self._k_models.keys()))

	def required_data(self):
		# combine all required_data from all class-level submodels
		req = Dict()
		for k_name, k_model in self._k_models.items():
			k_req = k_model.required_data()
			for i in ['ca','co']:
				if i in k_req or i in req:
					req[i] = list(set(req.get(i,[])) | set(k_req.get(i,[])))
			for i in ['weight_co', 'avail_ca', 'avail_co', 'choice_ca', 'choice_co_code', 'choice_co_vars', ]:
				if i in req:
					if i in k_req and req[i] != k_req[i]:
						raise ValueError(f'mismatched {i}')
					else:
						pass
				else:
					if i in k_req:
						req[i] = k_req[i]
					else:
						pass
		top_req = self._k_membership.required_data()
		if 'co' in top_req:
			req['co'] = list(sorted(set(req.get('co', [])) | set(top_req.get('co', []))))
		return req

	def total_weight(self):
		"""
		The total weight of cases in the attached dataframes.

		Returns
		-------
		float
		"""
		if self._dataframes is None:
			raise MissingDataError("no dataframes are set")
		return self._dataframes.total_weight()

	def __prep_for_compute(self, x=None):
		self.unmangle()
		if x is not None:
			self._k_membership.set_values(x)

	def class_membership_probability(self, x=None, start_case=0, stop_case=-1, step_case=1):
		self.__prep_for_compute(x)
		return self._k_membership.probability(
			x=None,
			return_dataframe='idco',
			start_case=start_case,
			stop_case=stop_case,
			step_case=step_case,
		)

	def class_membership_d_probability(self, x=None, start_case=0, stop_case=-1, step_case=1):
		self.__prep_for_compute(x)
		return self._k_membership.d_probability(
			x=None,
			start_case=start_case,
			stop_case=stop_case,
			step_case=step_case,
		)

	def probability(self, x=None, start_case=0, stop_case=-1, step_case=1, return_dataframe=False,):
		self.__prep_for_compute(x)

		if start_case >= self.dataframes.n_cases:
			raise IndexError("start_case >= n_cases")

		if stop_case == -1:
			stop_case = self.dataframes.n_cases

		if start_case > stop_case:
			raise IndexError("start_case > stop_case")

		if step_case <= 0:
			raise IndexError("non-positive step_case")

		n_rows = ((stop_case - start_case) // step_case) + (1 if (stop_case - start_case) % step_case else 0)

		p = numpy.zeros([n_rows, self.dataframes.n_alts])

		import warnings
		with warnings.catch_warnings():
			warnings.simplefilter("ignore", category=ParameterNotInModelWarning)
			k_membership_probability = self.class_membership_probability(
				start_case=start_case, stop_case=stop_case, step_case=step_case,
			)
			for k_name, k_model in self._k_models.items():
				k_pr = k_model.probability(start_case=start_case, stop_case=stop_case, step_case=step_case)
				p += (
						numpy.asarray( k_pr[:,:self.dataframes.n_alts] )
						* k_membership_probability.loc[:,k_name].values[:, None]
				)

		if return_dataframe:
			return pandas.DataFrame(
				p,
				index=self._dataframes.caseindex[start_case:stop_case:step_case],
				columns=self._dataframes.alternative_codes(),
			)
		return p


	def d_probability(self, x=None, start_case=0, stop_case=-1, step_case=1,):
		"""
		Compute the partial derivative of probability w.r.t. the parameters.

		Note this function is known to be incomplete.  It computes the
		derivative only within the classes, not for the class membership model.

		Parameters
		----------
		x
		start_case
		stop_case
		step_case
		return_dataframe

		Returns
		-------

		"""
		self.__prep_for_compute(x)

		if start_case >= self.dataframes.n_cases:
			raise IndexError("start_case >= n_cases")

		if stop_case == -1:
			stop_case = self.dataframes.n_cases

		if start_case > stop_case:
			raise IndexError("start_case > stop_case")

		if step_case <= 0:
			raise IndexError("non-positive step_case")

		n_rows = ((stop_case - start_case) // step_case) + (1 if (stop_case - start_case) % step_case else 0)

		p = numpy.zeros([n_rows, self.dataframes.n_alts, len(self.pf)])

		import warnings
		with warnings.catch_warnings():
			warnings.simplefilter("ignore", category=ParameterNotInModelWarning)
			k_membership_probability = self.class_membership_probability(
				start_case=start_case, stop_case=stop_case, step_case=step_case,
			)
			k_membership_d_probability = self.class_membership_d_probability(
				start_case=start_case, stop_case=stop_case, step_case=step_case,
			)
			for k_name, k_model in self._k_models.items():
				k_pr = k_model.probability(start_case=start_case, stop_case=stop_case, step_case=step_case)
				k_d_pr = k_model.d_probability(start_case=start_case, stop_case=stop_case, step_case=step_case)
				p += (
						numpy.asarray( k_d_pr[:,:self.dataframes.n_alts,:] )
						* k_membership_probability.loc[:,k_name].values[:, None, None]
				)
				k_position = k_membership_probability.columns.get_loc(k_name)
				p += (
						numpy.asarray( k_pr[:,:self.dataframes.n_alts, None] )
						* k_membership_d_probability[:,k_position,:][:,None,:]
				)
		return p


	def loglike2(
			self,
			x=None,
			*,
			start_case=0,
			stop_case=-1,
			step_case=1,
			persist=0,
			leave_out=-1,
			keep_only=-1,
			subsample=-1,
			return_series=True,
			probability_only=False,
	):
		"""
		Compute a log likelihood value and it first derivative.

		Parameters
		----------
		x : {'null', 'init', 'best', array-like, dict, scalar}, optional
			Values for the parameters.  See :ref:`set_values` for details.
		start_case : int, default 0
			The first case to include in the log likelihood computation.  To include all
			cases, start from 0 (the default).
		stop_case : int, default -1
			One past the last case to include in the log likelihood computation.  This is processed as usual for
			Python slicing and iterating, and negative values count backward from the end.  To include all cases,
			end at -1 (the default).
		step_case : int, default 1
			The step size of the case iterator to use in likelihood calculation.  This is processed as usual for
			Python slicing and iterating.  To include all cases, step by 1 (the default).
		persist : int, default 0
			Whether to return a variety of internal and intermediate arrays in the result dictionary.
			If set to 0, only the final `ll` value is included.
		leave_out, keep_only, subsample : int, optional
			Settings for cross validation calculations.
			If `leave_out` and `subsample` are set, then case rows where rownumber % subsample == leave_out are dropped.
			If `keep_only` and `subsample` are set, then only case rows where rownumber % subsample == keep_only are used.
		return_series : bool
			Deprecated, no effect.  Derivatives are always returned as a Series.

		Returns
		-------
		dictx
			The log likelihood is given by key 'll' and the first derivative by key 'dll'.
			Other arrays are also included if `persist` is set to True.

		"""
		if leave_out != -1 or keep_only != -1 or subsample != -1:
			raise NotImplementedError()

		from ..util import dictx

		self.__prep_for_compute(x)
		pr = self.probability(
			x=None,
			start_case=start_case, stop_case=stop_case, step_case=step_case,
		)

		y = dictx()

		if probability_only:
			y.ll = numpy.nan
			y.probability = pr
			return y

		d_p = self.d_probability(
			x=None,
			start_case=start_case, stop_case=stop_case, step_case=step_case,
		)

		from .nl import d_loglike_from_d_probability
		from .mnl import loglike_from_probability

		if stop_case == -1:
			stop_case_ = self.n_cases
		else:
			stop_case_ = stop_case

		if self.dataframes.data_wt is not None:
			wt = self.dataframes.data_wt.iloc[start_case:stop_case_:step_case]
		else:
			wt = None

		ch = self.dataframes.array_ch()[start_case:stop_case_:step_case]

		y.ll = loglike_from_probability(
			pr,
			ch,
			wt
		)

		if persist & persist_flags.PERSIST_PROBABILITY:
			y.probability = pr

		if persist & persist_flags.PERSIST_BHHH:
			y.dll, y.bhhh = d_loglike_from_d_probability(
				pr,
				d_p,
				ch,
				wt,
				True
			)
		else:
			y.dll = d_loglike_from_d_probability(
				pr,
				d_p,
				ch,
				wt,
				False
			)

		if start_case==0 and (stop_case==-1 or stop_case==self.n_cases) and step_case==1:
			self._check_if_best(y.ll)

		return y

	def loglike2_bhhh(
			self,
			x=None,
			*,
			start_case=0,
			stop_case=-1,
			step_case=1,
			persist=0,
			leave_out=-1,
			keep_only=-1,
			subsample=-1,
			return_series=False,
	):
		"""
		Compute a log likelihood value, it first derivative, and the BHHH approximation of the Hessian.

		The `BHHH algorithm <https://en.wikipedia.org/wiki/Berndt–Hall–Hall–Hausman_algorithm>`
		employs a matrix computated as the sum of the casewise outer product of the gradient, to
		approximate the hessian matrix.

		Parameters
		----------
		x : {'null', 'init', 'best', array-like, dict, scalar}, optional
			Values for the parameters.  See :ref:`set_values` for details.
		start_case : int, default 0
			The first case to include in the log likelihood computation.  To include all
			cases, start from 0 (the default).
		stop_case : int, default -1
			One past the last case to include in the log likelihood computation.  This is processed as usual for
			Python slicing and iterating, and negative values count backward from the end.  To include all cases,
			end at -1 (the default).
		step_case : int, default 1
			The step size of the case iterator to use in likelihood calculation.  This is processed as usual for
			Python slicing and iterating.  To include all cases, step by 1 (the default).
		persist : int, default False
			Whether to return a variety of internal and intermediate arrays in the result dictionary.
			If set to 0, only the final `ll` value is included.
		leave_out, keep_only, subsample : int, optional
			Settings for cross validation calculations.
			If `leave_out` and `subsample` are set, then case rows where rownumber % subsample == leave_out are dropped.
			If `keep_only` and `subsample` are set, then only case rows where rownumber % subsample == keep_only are used.
		return_series : bool
			Deprecated, no effect.  Derivatives are always returned as a Series.

		Returns
		-------
		dictx
			The log likelihood is given by key 'll', the first derivative by key 'dll', and the BHHH matrix by 'bhhh'.
			Other arrays are also included if `persist` is set to True.

		"""
		return self.loglike2(
			x=x,
			start_case=start_case,
			stop_case=stop_case,
			step_case=step_case,
			persist=persist | persist_flags.PERSIST_BHHH,
			leave_out=leave_out,
			keep_only=keep_only,
			subsample=subsample,
			return_series=True,
			probability_only=False,
		)

	def loglike(
			self,
			x=None,
			*,
			start_case=0, stop_case=-1, step_case=1,
			persist=0,
			leave_out=-1, keep_only=-1, subsample=-1,
			probability_only=False,
	):
		"""
		Compute a log likelihood value.

		Parameters
		----------
		x : {'null', 'init', 'best', array-like, dict, scalar}, optional
			Values for the parameters.  See :ref:`set_values` for details.
		start_case : int, default 0
			The first case to include in the log likelihood computation.  To include all
			cases, start from 0 (the default).
		stop_case : int, default -1
			One past the last case to include in the log likelihood computation.  This is processed as usual for
			Python slicing and iterating, and negative values count backward from the end.  To include all cases,
			end at -1 (the default).
		step_case : int, default 1
			The step size of the case iterator to use in likelihood calculation.  This is processed as usual for
			Python slicing and iterating.  To include all cases, step by 1 (the default).
		persist : int, default 0
			Whether to return a variety of internal and intermediate arrays in the result dictionary.
			If set to 0, only the final `ll` value is included.
		leave_out, keep_only, subsample : int, optional
			Settings for cross validation calculations.
			If `leave_out` and `subsample` are set, then case rows where rownumber % subsample == leave_out are dropped.
			If `keep_only` and `subsample` are set, then only case rows where rownumber % subsample == keep_only are used.
		probability_only : bool, default False
			Compute only the probability and ignore the likelihood.  If this setting is active, the
			dataframes need not include the "actual" choice data.

		Returns
		-------
		float or array or dictx
			The log likelihood as a float, when `probability_only` is False and `persist` is 0.
			The probability as an array, when `probability_only` is True and `persist` is 0.
			A dictx is returned if `persist` is non-zero.
		"""
		self.__prep_for_compute(x)
		pr = self.probability(
			x=None,
			start_case=start_case,
			stop_case=stop_case,
			step_case=step_case,
			return_dataframe=False,
		)
		if probability_only:
			return pr
		from .mnl import loglike_from_probability

		if stop_case == -1:
			stop_case_ = self.n_cases
		else:
			stop_case_ = stop_case

		if self.dataframes.data_wt is not None:
			wt = self.dataframes.data_wt.iloc[start_case:stop_case_:step_case]
		else:
			wt = None

		from ..util import dictx
		y = dictx()
		y.ll = loglike_from_probability(
			pr,
			self.dataframes.array_ch()[start_case:stop_case_:step_case],
			wt
		)

		if start_case==0 and (stop_case==-1 or stop_case==self.n_cases) and step_case==1:
			self._check_if_best(y.ll)

		if persist & persist_flags.PERSIST_PROBABILITY:
			y.probability = pr
		if persist:
			return y
		return y.ll



	@property
	def dataframes(self):
		return self._dataframes

	@dataframes.setter
	def dataframes(self, x):
		self._dataframes = x

		if self._dataframes.data_co is None and self._dataframes.data_ce is not None:
			self._dataframes.data_co = pandas.DataFrame(data=None, index=self._dataframes.data_ce.index.levels[0].copy())

		if self._dataframes.data_co is None and self._dataframes.data_ca is not None:
			self._dataframes.data_co = pandas.DataFrame(data=None, index=self._dataframes.data_ca.index.levels[0].copy())

		top_data = DataFrames(
			co = x.make_idco(*self._k_membership.required_data().get('co', [])),
			alt_codes=numpy.arange(1, len(self._k_models)+1),
			alt_names=self._k_model_names(),
			av=1,
		)

		# TODO: This kludge creates an empty array for the data_ch in the class membership model
		#     : but it is not needed except to satisfy a data integrity check on that model
		#     : We should instead just allow no-choice when it is a class membership model.
		# top_data.data_ch = pandas.DataFrame(0, index=top_data.caseindex, columns=self._k_models.keys())
		self._k_membership._model_does_not_require_choice = True

		self._k_membership.dataframes = top_data
		for k_name, k_model in self._k_models.items():
			k_model.dataframes = DataFrames(
				co=x.data_co,
				ca=x.data_ca,
				ce=x.data_ce,
				av=x.data_av,
				ch=x.data_ch,
				wt=x.data_wt,
				alt_names=x.alternative_names(),
				alt_codes=x.alternative_codes(),
			)

	def mangle(self, *args, **kwargs):
		self._k_membership.mangle()
		for m in self._k_models.values():
			m.mangle()
		super().mangle(*args, **kwargs)

	def unmangle(self, force=False):
		super().unmangle(force=force)
		if self._mangled or force:
			self._k_membership.unmangle(force=force)
			for m in self._k_models.values():
				m.unmangle(force=force)
			sync_frames(self, self._k_membership, *self._k_models.values())
			self._k_membership.unmangle()
			for m in self._k_models.values():
				m.unmangle()
			self._mangled = False

	def load_data(self, dataservice=None, autoscale_weights=True, log_warnings=True):
		self.unmangle()
		if dataservice is not None:
			self._dataservice = dataservice
		if self._dataservice is not None:
			dfs = self._dataservice.make_dataframes(self.required_data(), log_warnings=log_warnings)
			if autoscale_weights and dfs.data_wt is not None:
				dfs.autoscale_weights()
			self.dataframes = dfs
		else:
			raise ValueError('dataservice is not defined')

	def set_value(self, name, value=None, **kwargs):
		if isinstance(name, ParameterRef_C):
			name = str(name)
		if name not in self.pf.index:
			self.unmangle()
		if value is not None:
			# value = numpy.float64(value)
			# self.frame.loc[name,'value'] = value
			kwargs['value'] = value
		for k,v in kwargs.items():
			if k in self.pf.columns:
				if self.pf.dtypes[k] == 'float64':
					v = numpy.float64(v)
				elif self.pf.dtypes[k] == 'float32':
					v = numpy.float32(v)
				elif self.pf.dtypes[k] == 'int8':
					v = numpy.int8(v)
				self.pf.loc[name, k] = v

			# update init values when they are implied
			if k=='value':
				if 'initvalue' not in kwargs and pandas.isnull(self.pf.loc[name, 'initvalue']):
					self.pf.loc[name, 'initvalue'] = l4_float_dtype(v)

		# update null values when they are implied
		if 'nullvalue' not in kwargs and pandas.isnull(self.pf.loc[name, 'nullvalue']):
			self.pf.loc[name, 'nullvalue'] = 0

		# refresh everything # TODO: only refresh what changed
		self._k_membership._check_if_frame_values_changed()
		for m in self._k_models.values():
			m._check_if_frame_values_changed()

	def _frame_values_have_changed(self):
		self._k_membership._frame_values_have_changed()
		for m in self._k_models.values():
			m._frame_values_have_changed()

	@property
	def n_cases(self):
		"""int : The number of cases in the attached dataframes."""
		if self._dataframes is None:
			raise MissingDataError("no dataframes are set")
		return self._k_membership.n_cases






default_config = """---

rules:
  braces:
    min-spaces-inside: 0
    max-spaces-inside: 0
    min-spaces-inside-empty: -1
    max-spaces-inside-empty: -1
  brackets:
    min-spaces-inside: 0
    max-spaces-inside: 0
    min-spaces-inside-empty: -1
    max-spaces-inside-empty: -1
  colons:
    max-spaces-before: 0
    max-spaces-after: 999
  commas:
    max-spaces-before: 0
    min-spaces-after: 1
    max-spaces-after: 1
  comments:
    level: warning
    require-starting-space: false
    min-spaces-from-content: 1
  comments-indentation:
    level: warning
  document-end: disable
  document-start: disable
  #  level: warning
  #  present: true
  empty-lines:
    max: 99
    max-start: 99
    max-end: 99
  empty-values:
    forbid-in-block-mappings: false
    forbid-in-flow-mappings: false
  hyphens:
    max-spaces-after: 1
  indentation:
    spaces: consistent
    indent-sequences: true
    check-multi-line-strings: false
  key-duplicates: enable
  key-ordering: disable
  line-length:
    max: 880
    allow-non-breakable-words: true
    allow-non-breakable-inline-mappings: false
  new-line-at-end-of-file: enable
  new-lines:
    type: unix
  octal-values:
    forbid-implicit-octal: false
    forbid-explicit-octal: false
  trailing-spaces: disable
  truthy: disable
  #  level: warning
"""


class Format(object):
	@staticmethod
	def parsable(problem, filename):
		return ('%(file)s:%(line)s:%(column)s: [%(level)s] %(message)s' %
				{'file': filename,
				 'line': problem.line,
				 'column': problem.column,
				 'level': problem.level,
				 'message': problem.message})

	@staticmethod
	def standard(problem, filename):
		line = '  %d:%d' % (problem.line, problem.column)
		line += max(12 - len(line), 0) * ' '
		line += problem.level
		line += max(21 - len(line), 0) * ' '
		line += problem.desc
		if problem.rule:
			line += '  (%s)' % problem.rule
		return line

	@staticmethod
	def standard_color(problem, filename):
		line = '  \033[2m%d:%d\033[0m' % (problem.line, problem.column)
		line += max(20 - len(line), 0) * ' '
		if problem.level == 'warning':
			line += '\033[33m%s\033[0m' % problem.level
		else:
			line += '\033[31m%s\033[0m' % problem.level
		line += max(38 - len(line), 0) * ' '
		line += problem.desc
		if problem.rule:
			line += '  \033[2m(%s)\033[0m' % problem.rule
		return line


def yaml_check(file, config_file=None, logger=None, encoding='utf-8'):

	if logger is None:
		log = print
	else:
		log = logger.error


	try:
		from yamllint import linter
		from yamllint.config import YamlLintConfig
		from yamllint.linter import PROBLEM_LEVELS
	except ImportError:
		log("yamllint is not installed, cannot inspect yaml file for formatting quality")
	else:

		filepath = file[2:] if file.startswith('./') else file

		if config_file is not None:
			conf = YamlLintConfig(file=config_file)
		else:
			conf = YamlLintConfig(content=default_config)

		first = True
		any_errors = False
		max_level = 0
		with open(filepath, 'r', encoding=encoding) as f:
			for problem in linter.run(f, conf, filepath):
				if first:
					log(f"FOUND YAML ERRORS IN {file}")
					first = False
					any_errors = True

				log(Format.standard(problem, file))

				max_level = max(max_level, PROBLEM_LEVELS[problem.level])

			if any_errors:
				log(f"END OF YAML ERRORS IN {file}")



import larch
import pandas as pd
from larch import P,X,PX
from pytest import approx
from larch.data_warehouse import example_file

def test_simple_model_group():

	df = pd.read_csv(example_file("MTCwork.csv.gz"))
	df.set_index(['casenum','altnum'], inplace=True)
	d = larch.DataFrames(df, ch='chose', crack=True)
	d.set_alternative_names({
		1: 'DA',
		2: 'SR2',
		3: 'SR3+',
		4: 'Transit',
		5: 'Bike',
		6: 'Walk',
	})

	m0 = larch.Model(dataservice=d)
	m0.utility_co[2] = P("ASC_SR2")  + P("hhinc#2") * X("hhinc")
	m0.utility_co[3] = P("ASC_SR3P") + P("hhinc#3") * X("hhinc")
	m0.utility_co[4] = P("ASC_TRAN") + P("hhinc#4") * X("hhinc")
	m0.utility_co[5] = P("ASC_BIKE") + P("hhinc#5") * X("hhinc")
	m0.utility_co[6] = P("ASC_WALK") + P("hhinc#6") * X("hhinc")
	m0.utility_ca = (
		(P("tottime_m")*X("tottime") + P("totcost_m")*X("totcost"))*X("femdum == 0")
		+
		(P("tottime_f")*X("tottime") + P("totcost_f")*X("totcost"))*X("femdum == 1")
	)

	m1 = larch.Model(dataservice=d.selector_co("femdum == 0"))
	m1.utility_co[2] = P("ASC_SR2")  + P("hhinc#2") * X("hhinc")
	m1.utility_co[3] = P("ASC_SR3P") + P("hhinc#3") * X("hhinc")
	m1.utility_co[4] = P("ASC_TRAN") + P("hhinc#4") * X("hhinc")
	m1.utility_co[5] = P("ASC_BIKE") + P("hhinc#5") * X("hhinc")
	m1.utility_co[6] = P("ASC_WALK") + P("hhinc#6") * X("hhinc")
	m1.utility_ca = P("tottime_m")*X("tottime") + P("totcost_m")*X("totcost")

	m2 = larch.Model(dataservice=d.selector_co("femdum == 1"))
	m2.utility_co[2] = P("ASC_SR2")  + P("hhinc#2") * X("hhinc")
	m2.utility_co[3] = P("ASC_SR3P") + P("hhinc#3") * X("hhinc")
	m2.utility_co[4] = P("ASC_TRAN") + P("hhinc#4") * X("hhinc")
	m2.utility_co[5] = P("ASC_BIKE") + P("hhinc#5") * X("hhinc")
	m2.utility_co[6] = P("ASC_WALK") + P("hhinc#6") * X("hhinc")
	m2.utility_ca = P("tottime_f")*X("tottime") + P("totcost_f")*X("totcost")

	m0.load_data()
	assert m0.loglike2().ll == approx(-7309.600971749625)

	m1.load_data()
	assert m1.loglike2().ll == approx(-4068.8091617468717)

	m2.load_data()
	assert m2.loglike2().ll == approx(-3240.7918100027578)

	from larch.model.model_group import ModelGroup

	mg = ModelGroup([m1,m2])

	assert mg.loglike2().ll == approx(-7309.600971749625)
	assert mg.loglike() == approx(-7309.600971749625)

	pd.testing.assert_series_equal(
		mg.loglike2().dll.sort_index(), m0.loglike2().dll.sort_index()
	)

	m0.simple_step_bhhh()
	mg.set_values(**m0.pf.value)

	pd.testing.assert_series_equal(
		mg.loglike2().dll.sort_index(), m0.loglike2().dll.sort_index()
	)

	assert mg.loglike2().ll == approx(-4926.4822036792275)
	assert mg.check_d_loglike().data.similarity.min() > 4

	result = mg.maximize_loglike(method='slsqp')
	assert result.loglike == approx(-3620.697668335103)

	mg2 = ModelGroup([])
	mg2.append(m1)
	mg2.append(m2)
	assert mg2.loglike() == approx(-3620.697667552756)




__version__ = '5.4.6'

from .util.interface_info import Info, ipython_status
import sys
from .util.styles import css

info = Info('Larch', False, __version__)
version = Info('Larch', False, __version__, minimal=True)

if 'IPython' in ipython_status():
    from .util.display import display
    try:
        # from .util.styles import stylesheet
        # stylesheet()
        # display(info)
        pass
    except:
        # print(repr(info))
        jupyter_active = False
    else:
        jupyter_active = True
else:
    jupyter_active = False
    # print(repr(info))


from .roles import P, X, PX
from .data_services import DataService
from .omx import OMX
from .data_services.dbf.dbf_reader import DBF

from .model import Model
from .dataframes import DataFrames

from .examples import example
from .util import figures
from .util.excel import ExcelWriter

_doctest_mode_ = False

_larch_self = sys.modules[__name__]

from xmle import Reporter, NumberedCaption
from xmle import load_metadata as read_metadata
from .workspace import make_reporter




from .generic import *
from tables import NoSuchNodeError

class H5PodGA(H5Pod):
	"""
	A HDF5 group node containing :ref:`idga` format data.

	The GA pod is conceptually similar to the CA pod, although cases are included
	by a lookup key instead of individually.  Each array of data is stored as an array, where
	each array has rows of values attributable to arbitrary groups of casewise observations,
	and columns attributable to individual choice alternatives.  The common use case is an OMX
	matrix file containing skims, where the case-wise observations each pull a single row
	from the skims.

	The H5PodGA has one special attribute , naming the row-index to use for each case.  This
	name must exist in :ref:`idco` data.

	"""

	def __init__(self, filename, rowindexes, *args, **kwargs):
		super().__init__(*args, filename=filename, **kwargs)
		if isinstance(rowindexes, str):
			self.rowindexes = self._groupnode._v_children[rowindexes][:]
		elif isinstance(rowindexes, numpy.ndarray):
			self.rowindexes = rowindexes.squeeze()
		else:
			raise TypeError('rowindexes must be str or ndarray')

	@property
	def podtype(self):
		return 'idga'


	def __getitem__(self, item):

		if isinstance(item, tuple) and len(item)>=2 and isinstance(item[-1], slice):
			names, slice_ = item[:-1], item[-1]
		else:
			names = item
			slice_ = None

		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names, ]

		dtype = numpy.float64

		result = numpy.zeros( [selector_len_for(slice_, self.shape[0]), *self.shape[1:], len(names)], dtype=dtype)

		for i, cmd in enumerate(names):
			temp = self._evaluate_single_item(cmd, None)
			if slice_ is not None:
				result[...,i] = temp[self.rowindexes[slice_],:]
			else:
				result[...,i] = temp[self.rowindexes,:]
		return result

	@property
	def shape(self):
		"""The shape of the pod.

		"""
		return tuple(self.rowindexes.shape) + self.metashape[1:]

	@property
	def metashape(self):
		"""The shape of the underlying skims.

		"""
		try:
			return tuple(self._groupnode._v_attrs.SHAPE)
		except (AttributeError, NoSuchNodeError):
			return tuple(self._groupnode._v_parent._v_attrs.SHAPE)

	def load_data_item(self, name, result, selector=None):
		"""Load a slice of the pod arrays into an array in memory"""
		# convert a single name string to a one item list

		from ...general import _sqz_same
		_sqz_same(result.shape, [selector_len_for(selector, self.shape[0]), *self.shape[1:]])

		temp = self._evaluate_single_item(name, None)
		if selector is not None:
			result[:] = temp[self.rowindexes[selector], :]
		else:
			result[:] = temp[self.rowindexes, :]
		return result





from xmle import Reporter, NumberedCaption

def make_reporter(title, notes=True, viz=True):
	"""
	Make a standard reporter.
	"""

	REPORT = Reporter(title)
	REPORT.MAIN = REPORT.section("Estimated Model", "Estimated Model")
	if notes is True:
		REPORT.NOTES = REPORT.section("Model Notes", "Notes")
	elif notes:
		REPORT.NOTES = REPORT.section(notes, notes)
	if viz is True:
		REPORT.VIZ = REPORT.section("Model Visualization", "Visualization")
	elif viz:
		REPORT.VIZ = REPORT.section(viz, viz)

	REPORT.FIG = NumberedCaption('Figure')
	REPORT.TAB = NumberedCaption('Table')

	return REPORT




# -*- coding: utf-8 -*-

import appdirs
import joblib
import requests

cache_dir = None
memory = None

def set_cache_dir(location=None, compress=True, verbose=0, **kwargs):
	"""
	Set up a cache directory for use with requests.

	Parameters
	----------
	location: str or None or False
		The path of the base directory to use as a data store
		or None or False.  If None, a default directory is created
		using appdirs.user_cache_dir.
		If False is given, no caching is done and
		the Memory object is completely transparent.

	compress: boolean, or integer, optional
		Whether to zip the stored data on disk. If an integer is
		given, it should be between 1 and 9, and sets the amount
		of compression.

	verbose: int, optional
		Verbosity flag, controls the debug messages that are issued
		as functions are evaluated.

	bytes_limit: int, optional
		Limit in bytes of the size of the cache.

	"""
	global memory, cache_dir

	if location is None:
		location = appdirs.user_cache_dir('cached_requests')

	if location is False:
		location = None

	memory = joblib.Memory(location, compress=compress, verbose=verbose, **kwargs)

	make_cache = (
		(requests, 'get'),
		(requests, 'post'),
	)

	for module, func_name in make_cache:
		try:
			func = getattr(module, f"_{func_name}_orig")
		except AttributeError:
			func = getattr(module, func_name)
			setattr(module, f"_{func_name}_orig", func)
		setattr(module, func_name, memory.cache(func))





set_cache_dir()






import numpy
import numpy.ma as ma
from .colors import hexcolor

from numpy.lib.histograms import _hist_bin_selectors


def robust_min_max(s, threshold_for_dropping_zeros=0.35, cut_after_stdevs=5, percentiles_to_cut=1):
	# Get s as an array (it should be already though)
	s = numpy.asarray(s)
	# mask out infs and nans
	sx = ma.masked_array(s, mask=~numpy.isfinite(s))
	n_values = sx.count()
	# Check if data is mostly zeros
	n_zeros = (s==0).sum()
	if n_zeros/n_values > threshold_for_dropping_zeros:
		sx.mask |= (sx == 0)
		zeros_are_dropped = True
		n_values = sx.count()
	else:
		zeros_are_dropped = False
	if n_values==0:
		# no data is left, default values
		bottom, top, range_is_truncated = 0, 1, 100
	else:
		mean = sx.mean()
		stdev = sx.std()
		bottom = sx.min()
		top = sx.max()
		range_is_truncated = 0
		cut_bottom = (bottom < mean - cut_after_stdevs * stdev)
		cut_top = (top > mean + cut_after_stdevs * stdev)
		if percentiles_to_cut>0:
			if cut_bottom:
				bottom_ = numpy.percentile(sx, percentiles_to_cut)
				if not numpy.isnan(bottom_):
					bottom = bottom_
					range_is_truncated += percentiles_to_cut
			if cut_top:
				top_ = numpy.percentile(sx, 100-percentiles_to_cut)
				if not numpy.isnan(top_):
					top = top_
					range_is_truncated += percentiles_to_cut
	return bottom, top, zeros_are_dropped, range_is_truncated, sx


def fixed_min_max(s, fixed_min, fixed_max, threshold_for_dropping_zeros=0.35):
	# Get s as an array (it should be already though)
	s = numpy.asarray(s)
	# mask out infs and nans
	sx = ma.masked_array(s, mask=numpy.isnan(s))
	n_values_finite = n_values = sx.count()

	n_zeros = (s==0).sum()
	if n_zeros/n_values > threshold_for_dropping_zeros:
		sx.mask |= (sx == 0)
		zeros_are_dropped = True
		n_values = sx.count()
	else:
		zeros_are_dropped = False

	# Over max
	if fixed_max is not None:
		sx.mask |= (sx >= fixed_max)
	n_values_overmax = n_values - sx.count()
	n_values = sx.count()

	# Under Min
	if fixed_min is not None:
		sx.mask |= (sx <= fixed_min)
	n_values_undermin = n_values - sx.count()
	n_values = sx.count()

	pct_over = n_values_overmax / n_values_finite
	pct_under = n_values_undermin / n_values_finite

	return sx, pct_under, pct_over, zeros_are_dropped


def get_histogram_bins(data_to_bin, bins='sturges', percentiles_to_cut=1, maxrange=None, fixedrange=None, threshold_for_dropping_zeros=0.35):
	if fixedrange is None:
		bottom, top, zeros_are_dropped, range_is_truncated, sx = robust_min_max(data_to_bin, percentiles_to_cut=percentiles_to_cut)
		pct_under, pct_over = 0, 0
	else:
		bottom, top = fixedrange[0], fixedrange[1]
		sx, pct_under, pct_over, zeros_are_dropped = fixed_min_max(data_to_bin, bottom, top, threshold_for_dropping_zeros=threshold_for_dropping_zeros)
		range_is_truncated = (pct_under + pct_over)*100
	bottom = float(bottom) if bottom is not None else sx.min()-1e-8
	top = float(top) if top is not None else sx.max()+1e-8

	if numpy.isinf(bottom):
		bottom = sx.min()
	if numpy.isinf(top):
		top = sx.max()

	if maxrange is not None:
		if maxrange[0] > bottom:
			bottom = float(maxrange[0])
		if maxrange[1] < top:
			top = float(maxrange[1])

	if bottom > top: # unclear how this would happen, but just in case
		bottom, top = top, bottom

	# generate number of bins based on named algorithm
	if isinstance(bins, str):
		try:
			width = _hist_bin_selectors[bins](sx, (bottom, top))
		except IndexError:
			width = 1
		if numpy.isnan(width):
			width = 1
		try:
			if width:
				bins = int(numpy.ceil((top - bottom) / width))
			else:
				bins = 1
		except OverflowError:
			bins = 10
		# The spark graphs get hard to read if the bin slices are too thin, so we will max out at 50 bins
		if bins > 50:
			bins = 50
		# The spark graphs look weird if the bin slices are too fat, so we will min at 10 bins
		if bins < 10:
			bins = 10

	# generate actual bins based on integer number of bins
	if isinstance(bins, int):
		try:
			bins = numpy.linspace(bottom, top, num=bins, endpoint=True, retstep=False, dtype=None)
		except UnboundLocalError:
			raise

	return bins, bottom, top, zeros_are_dropped, range_is_truncated, sx, pct_under, pct_over


from numpy.ma.core import nomask as _nomask
import pandas

def _compress_by_mask(arr, mask=None):
	if arr is None:
		return None
	if isinstance(arr, numpy.ma.core.MaskedArray):
		if mask is None:
			mask = arr
		arr = arr._data
	if isinstance(arr, pandas.Series):
		arr = arr.data
	if isinstance(arr, memoryview):
		arr = numpy.asarray(arr)
	data = numpy.ndarray.ravel(arr)
	if mask._mask is not _nomask:
		data = data.compress(numpy.logical_not(numpy.ndarray.ravel(mask._mask)))
	return data


def histogram_data(
		data_to_bin,
		bins='sturges',
		weights=None,
		ch_weights=None,
		density=None,
		percentiles_to_cut=1,
		maxrange=None,
		fixedrange=None,
		threshold_for_dropping_zeros=0.35
):
	bins, bottom, top, zeros_are_dropped, range_is_truncated, sx, pct_under, pct_over = get_histogram_bins(
		data_to_bin, bins=bins, percentiles_to_cut=percentiles_to_cut, maxrange=maxrange, fixedrange=fixedrange, threshold_for_dropping_zeros=threshold_for_dropping_zeros,
	)

	h, bin_edges = numpy.histogram(_compress_by_mask(sx), bins=bins, range=(bottom, top), weights=_compress_by_mask(weights,sx), density=False)
	if ch_weights is not None:
		ch_h, _ = numpy.histogram(_compress_by_mask(sx), bins=bin_edges, weights=_compress_by_mask(ch_weights,sx), density=False, range=(bottom, top))
	else:
		ch_h = None

	if density:
		if fixedrange is not None and ch_weights is not None:
			# Get adjusted density for choice heights
			db = numpy.array(numpy.diff(bin_edges), float)
			if weights is None:
				h = h / db / sx._data.size
			else:
				h = h / db / weights.sum()
			ch_h = ch_h / db / ch_weights.sum()
		else:
			db = numpy.array(numpy.diff(bin_edges), float)
			h = h / db / h.sum()
			ch_h = ch_h / db / ch_h.sum()


	return h, bin_edges, zeros_are_dropped, range_is_truncated, ch_h, pct_under, pct_over


_histogram_notes = {
	hexcolor('orange'): "Histograms are orange if the zeros are numerous and have been excluded.",
	hexcolor('red'):    "Histograms are red if the zeros are numerous and have been excluded, and the displayed range truncates some extreme outliers.",
	hexcolor('forest'): "Histograms are green if the displayed range truncates some extreme outliers.",
	hexcolor('ocean'):  None,
	hexcolor('night'):  "Histograms are purple if the data is represented as discrete values.",
}

def draw_histogram_figure(
		bin_heights, bin_edges, zeros_are_dropped, range_is_truncated,
		title=None, xlabel=None, ylabel=None,
		xticks=False, yticks=False,
		xticklabels=None, yticklabels=False,
		frame_off='all', notetaker=None, bgcolor=None,
		figwidth=0.75, figheight=0.2,
		return_format='svg',
		attach_metadata=True,
		ch_heights=None,
		discrete=False,
		left_thermo=None,
		right_thermo=None,
	):
	from matplotlib import pyplot as plt
	fig, ax = plt.subplots(nrows=1, ncols=1, sharex=False, sharey=False)
	widths = bin_edges[1:] - bin_edges[:-1]
	if discrete:
		widths = widths * 0.8
		color = hexcolor('night')
	elif zeros_are_dropped:
		if range_is_truncated and not left_thermo and not right_thermo:
			color = hexcolor('red')
		else:
			color = hexcolor('orange')
	elif range_is_truncated and not left_thermo and not right_thermo:
		color = hexcolor('forest')
	else:
		color = hexcolor('ocean')
	rects1 = ax.bar(
		bin_edges[:-1],
		bin_heights,
		widths,
		color=color,
		align= 'center' if discrete else 'edge'
	)

	chart_top = bin_heights.max()

	if ch_heights is not None:
		rects2 = ax.bar(
			bin_edges[:-1],
			height=numpy.zeros_like(ch_heights),
			width=widths,
			bottom=ch_heights,
			edgecolor='black',
			linewidth=1.0,
			align='center' if discrete else 'edge',
			clip_on=False
		)
		chart_top = max(chart_top, ch_heights.max())

	if left_thermo is not None:
		leftmost, rightmost = bin_edges[0], bin_edges[-1]
		full_width = rightmost-leftmost
		ax.bar(
			numpy.asarray([leftmost - full_width * 0.05,]),
			height=chart_top * left_thermo,
			width=numpy.asarray([full_width * 0.035, ]),
			color='red',
			align='edge'
		)
		ax.bar(
			numpy.asarray([leftmost - full_width * 0.05,]),
			height=chart_top * (1-left_thermo),
			width=numpy.asarray([full_width * 0.035, ]),
			bottom=chart_top * left_thermo,
			color='#e9edd3',
			align='edge'
		)
		# Thermo-border
		ax.bar(
			numpy.asarray([leftmost - full_width * 0.05,]),
			height=chart_top,
			width=numpy.asarray([full_width * 0.035, ]),
			color='#FFFFFF00',
			align='edge',
			edgecolor = 'black',
			linewidth = 0.5,
			clip_on=False
		)

	if right_thermo is not None:
		leftmost, rightmost = bin_edges[0], bin_edges[-1]
		full_width = rightmost-leftmost
		ax.bar(
			numpy.asarray([rightmost + full_width * 0.015,]),
			height=chart_top * right_thermo,
			width=numpy.asarray([full_width * 0.035, ]),
			color='red',
			align='edge'
		)
		ax.bar(
			numpy.asarray([rightmost + full_width * 0.015,]),
			height=chart_top * (1-right_thermo),
			width=numpy.asarray([full_width * 0.035, ]),
			bottom=chart_top * right_thermo,
			color='#e9edd3',
			align='edge'
		)
		# Thermo-border
		ax.bar(
			numpy.asarray([rightmost + full_width * 0.015,]),
			height=chart_top,
			width=numpy.asarray([full_width * 0.035, ]),
			color='#FFFFFF00',
			align='edge',
			edgecolor = 'black',
			linewidth = 0.5,
			clip_on=False
		)


	if figheight:
		fig.set_figheight(figheight)
	if figwidth:
		fig.set_figwidth(figwidth)
	fig.set_dpi(300)

	if frame_off == 'all' or frame_off is True or frame_off==1:
		frame_off = {'left', 'right', 'top', 'bottom'}
	elif frame_off is None:
		frame_off = set()
	elif not isinstance(frame_off, set):
		print(type(frame_off))
		print(frame_off)
		frame_off = set(frame_off)

	if xlabel: ax.set_xlabel(xlabel)
	if ylabel: ax.set_ylabel(ylabel)
	if title: ax.set_title(title)
	if not isinstance(xticks, (list,tuple,numpy.ndarray)) and not xticks:
		ax.get_xaxis().set_ticks([])
	else:
		frame_off.discard('bottom')
		if xticklabels is not None:
			ax.get_xaxis().set_ticklabels(xticklabels)
			ax.get_xaxis().set_ticks(xticks)
	if not isinstance(yticks, (list,tuple,numpy.ndarray)) and not yticks:
		ax.get_yaxis().set_ticks([])
	else:
		frame_off.discard('left')
		if yticklabels is not None:
			ax.get_yaxis().set_ticklabels(yticklabels)
			ax.get_yaxis().set_ticks(yticks)

	if frame_off is not None:
		if isinstance(frame_off, str):
			frame_off = [frame_off]
		from .plotting import adjust_spines
		sides = {'left', 'right', 'top', 'bottom'}
		for side in frame_off:
			sides.remove(side)
		adjust_spines(ax, sides)

	if bgcolor is not None:
		ax.set_axis_bgcolor(bgcolor)

	if notetaker is not None and color != hexcolor('ocean'):
		notetaker.add(_histogram_notes[color])

	if return_format.lower() == 'svg':
		from .plotting import plot_as_svg_xhtml
		pad_inches = 0.1
		try:
			if 'left' in frame_off and 'bottom' in frame_off:
				ax.set_position([0,0,1,1], which='both')
				ax.set_frame_on(False)
				pad_inches=0
			elif figheight<=1.0:
				plt.tight_layout(pad=figheight/2)
				pad_inches=0
			else:
				plt.tight_layout(pad=0.5)
		except ValueError:
			pass # tight layout sometimes fails when the figsize is too small
		tooltip = _histogram_notes[color] or ""
		if ch_heights is not None:
			if len(tooltip):
				tooltip += "\n"
			tooltip += "Black marks indicate frequencies considering only chosen alternatives."
		if left_thermo or right_thermo:
			if len(tooltip):
				tooltip += "\n"
			tooltip += "Side thermometers indicate the fraction of all observations at or beyond the limits, and are scaled independently."

		ret = plot_as_svg_xhtml(plt, tooltip=tooltip, pad_inches=pad_inches)
		plt.clf()
		plt.close()  # do not showing empty windows?
		if attach_metadata:
			ret.metadata = dict(
				bin_heights=bin_heights,
				bin_edges=bin_edges,
				zeros_are_dropped=zeros_are_dropped,
				range_is_truncated=range_is_truncated)
			ret.attrib['zeros_are_dropped'] = str(zeros_are_dropped)
			ret.attrib['range_is_truncated'] = str(range_is_truncated)
		ret.attrib['style'] = f"min-width:{figwidth}in"
		return ret

def histogram_figure(
		data_to_bin,
		bins='sturges',
		weights=None,
		ch_weights=None,
		return_all=False,
		percentiles_to_cut=1,
		maxrange=None,
		fixedrange=None,
		piecerange=None,
		mask=None,
		**kwargs
):
	density = None if ch_weights is None else True
	discrete = (bins=='discrete')

	if mask is not None:
		data_to_bin = data_to_bin[mask]
		if weights is not None:
			weights = weights[mask]
		if ch_weights is not None:
			ch_weights = ch_weights[mask]

	threshold_for_dropping_zeros = 0.35

	if piecerange is not None and piecerange != (None,None):
		if piecerange[0] is None:
			fixedrange = (None, piecerange[1])
		elif piecerange[1] is None:
			fixedrange = (0, data_to_bin.max()+1e-8)
			threshold_for_dropping_zeros = 100
		else:
			fixedrange = (0, piecerange[1]-piecerange[0])
			threshold_for_dropping_zeros = 100
		percentiles_to_cut = 0
		discrete = False

	if discrete:
		bin_labels, bin_inv = numpy.unique(data_to_bin, return_inverse = True)
		bin_heights = numpy.bincount(bin_inv, weights)
		bin_edges = numpy.arange(len(bin_heights)+1)
		zeros_are_dropped, range_is_truncated = False, False
		if ch_weights is not None:
			ch_heights = numpy.bincount(bin_inv, ch_weights)
		else:
			ch_heights = None
		if density:
			bin_heights = bin_heights/(bin_heights.sum())
			if ch_heights is not None:
				ch_heights = ch_heights/(ch_heights.sum())
		if (('xticks' not in kwargs) and (kwargs.get('figwidth',1)>=2)) or (kwargs.get('xticks') is True):
			kwargs['xticklabels'] = bin_labels
			kwargs['xticks'] = bin_edges[:-1]
		pct_under, pct_over = 0,0
	else:
		bin_heights, bin_edges, zeros_are_dropped, range_is_truncated, ch_heights, pct_under, pct_over = histogram_data(
			data_to_bin, bins, weights, ch_weights, density=density, percentiles_to_cut=percentiles_to_cut, maxrange=maxrange, fixedrange=fixedrange, threshold_for_dropping_zeros=threshold_for_dropping_zeros
		)
		bin_labels = None

	if piecerange is not None and piecerange[0] is not None:
		bin_edges += piecerange[0]

	try:
		f = draw_histogram_figure(
			bin_heights, bin_edges, zeros_are_dropped, range_is_truncated,
			ch_heights = ch_heights,
			discrete=discrete,
			left_thermo = pct_under or None,
			right_thermo=pct_over or None,

			**kwargs
		)
	except:
		print("bin_heights\n",bin_heights)
		print("bin_edges\n", bin_edges)
		print("zeros_are_dropped\n", zeros_are_dropped)
		print("range_is_truncated\n", range_is_truncated)
		raise
	if return_all:
		return f, bin_heights, bin_edges, zeros_are_dropped, range_is_truncated
	else:
		return f


def sizable_histogram_figure(*args, sizer=1, discrete=None, on_error='ignore', **kwargs):
	kwargs['figwidth'] = sizer
	kwargs['figheight'] = sizer / 3.5
	if sizer >= 2:
		kwargs['frame_off'] = {'top', 'right'}
		kwargs['xticks'] = True
		#kwargs['yticks'] = True
	if sizer < 4:
		kwargs.pop('title', None)
	if sizer < 3:
		kwargs.pop('xlabel', None)
		kwargs.pop('ylabel', None)
	if discrete is not None:
		if discrete:
			kwargs['bins'] = 'discrete'
	try:
		return histogram_figure(*args, **kwargs)
	except:
		if on_error == 'ignore':
			return None
		else:
			raise

def seems_like_discrete_data(arr, dictionary=None, return_uniques=False):
	if isinstance(arr.dtype, pandas.CategoricalDtype):
		#print('seems_like_discrete_data? YES categorical')
		if return_uniques:
			return (True, list(arr.cat.categories))
		return True
	elif numpy.issubdtype(arr.dtype, numpy.bool_):
		#print('seems_like_discrete_data? YES bool')
		if return_uniques:
			return (True, [False, True])
		return True
	else:
		pass
		#print('seems_like_discrete_data? not bool but',arr.dtype)
	if dictionary is None:
		if len(numpy.unique(arr[:100]))<6:
			if len(numpy.unique(arr[:1000])) < 6:
				uniques = numpy.unique(arr)
				if len(uniques) < 6:
					#print('seems_like_discrete_data? YES uniques < 6')
					if return_uniques:
						return (True, uniques)
					return True
		#print('seems_like_discrete_data? too many and no dictionary')
	else:
		uniq = numpy.unique(arr)
		not_in_dict = 0
		for i in uniq:
			if i not in dictionary:
				not_in_dict += 1
		if not_in_dict > 2:
			#print(f'seems_like_discrete_data? dictionary but {not_in_dict} missing keys')
			if return_uniques:
				return False, None
			return False
		else:
			#print(f'seems_like_discrete_data? dictionary with {not_in_dict} missing keys')
			if return_uniques:
				return True, uniq
			return True
	if return_uniques:
		return False, None
	return False



import numpy


class NonUniqueLookup(Exception):
	pass


def is_sorted_and_unique(arr):
	arr = arr.reshape(-1)
	return numpy.all(arr[:-1 ] <arr[1:])

def is_sorted(arr):
	arr = arr.reshape(-1)
	return numpy.all(arr[:-1 ] <=arr[1:])

def label_to_index(labels, arr):
	"""Convert an array of lookup-able values into indexes.

	If you have an array of lookup-able values (e.g., TAZ identifiers) and you
	want to convert them to 0-based indexes for use in accessing matrix data,
	this is the function for you.

	Parameters
	----------
	labels : 1d array-like
		An array of labels.
	arr : array-like
		An array of values that appear in the label array. This method uses
		numpy.digitize to process values, so any target value that appears in `arr` but does
		not appear in the labels will be assigned the index of the smallest label
		value that is greater than the target, or the maximum label value if no label value
		is greater than the target.

	Returns
	-------
	array
		An array of index (int) values, with the same shape as `arr`.

	Raises
	------
	OMXNonUniqueLookup
		When the lookup does not contain a set of unique values, this tool is not appropriate.

	"""
	try:
		if len(labels ) ==0:
			return labels
	except TypeError:
		return labels
	labels = numpy.asarray(labels)
	if is_sorted_and_unique(labels):
		return numpy.digitize(arr, labels, right=True)
	uniq_labels, uniq_indexes = numpy.unique(labels, return_inverse=True)
	if len(uniq_labels) != len(labels):
		raise NonUniqueLookup("lookup '{}' does not have unique labels for each item".format(lookupname))
	index_malordered = numpy.digitize(arr, uniq_labels, right=True)
	return uniq_indexes[index_malordered]



def labels_to_unique_ids(various_labels, lowest_label_number=1):
	"""Convert an array or list of various labels to unique sequential integer code numbers.

	Parameters
	----------
	various_labels : array or list
		A bunch of various labels to re-map.  Duplicates are allowed.
	lowest_label_number : int
		The lowest label number returned, by default 1.

	Returns
	-------
	array
		An array of unique labels, in order.
	ids
		An array of integer code numbers, with the same shape as the
		input `various_labels`.

	"""
	ordered_labels, label_numbers = numpy.unique( various_labels, return_inverse=True )
	label_numbers += lowest_label_number
	return ordered_labels, label_numbers



def orthogonal_unit_vector(nDims, positionOne, dtype=numpy.float64):
	z = numpy.zeros(nDims, dtype=dtype)
	z[positionOne] = 1
	return z




def is_all_integer(arr):
	if arr.size < 100:
		if numpy.all(numpy.equal(numpy.mod(arr, 1), 0)):
			return True
	else:
		if numpy.all(numpy.equal(numpy.mod(arr.ravel()[:100], 1), 0)):
			if numpy.all(numpy.equal(numpy.mod(arr, 1), 0)):
				return True
	return False




def is_all_integer_or_nan(arr):
	arr_nan = arr[~numpy.isnan(arr)]
	if arr_nan.size < 100:
		if numpy.all(numpy.equal(numpy.mod(arr_nan, 1), 0)):
			return True
	else:
		if numpy.all(numpy.equal(numpy.mod(arr_nan.ravel()[:100], 1), 0)):
			if numpy.all(numpy.equal(numpy.mod(arr_nan, 1), 0)):
				return True
	return False




def convert_float_to_int_if_lossless(arr, inttype=numpy.int32):
	if arr.size < 100:
		if numpy.all(numpy.equal(numpy.mod(arr, 1), 0)):
			return arr.astype(inttype)
	else:
		if numpy.all(numpy.equal(numpy.mod(arr.ravel()[:100], 1), 0)):
			if numpy.all(numpy.equal(numpy.mod(arr, 1), 0)):
				return arr.astype(inttype)
	return arr


def failable_iter_to_set(iterable, transformer):
	s = set()
	for i in iterable:
		try:
			s.add(transformer(i))
		except AttributeError:
			pass
	return s

def failable_iter_to_unique(iterable, transformer):
	s_cache = None
	for i in iterable:
		try:
			s = transformer(i)
		except AttributeError:
			pass
		else:
			if s_cache is not None:
				if s_cache!= s:
					return None
			else:
				s_cache = s
	return s_cache


def unique_successful_transform(iterable, transformer, accept_longest=False):
	s = failable_iter_to_unique(iterable, transformer)
	if len(s) == 1:
		return s.pop()
	if accept_longest:
		candidate = None
		candidate_len = 0
		for i in s:
			if len(str(i)) > candidate_len:
				candidate_len = len(str(i))
				candidate = i
		return candidate
	return None


def scalarize(a):
	try:
		return a.item()
	except (ValueError, AttributeError):
		return a




import numpy as np
import matplotlib
from matplotlib.testing.decorators import image_comparison
import matplotlib.pyplot as plt

# @image_comparison(baseline_images=['histogram'],
#                   extensions=['png'])
# def test_spines_axes_positions():
# 	# SF bug 2852168
# 	fig = plt.figure()
# 	x = np.linspace(0,2*np.pi,100)
# 	y = 2*np.sin(x)
# 	ax = fig.add_subplot(1,1,1)
# 	ax.set_title('centered spines')
# 	ax.plot(x,y)
# 	ax.spines['right'].set_position(('axes',0.1))
# 	ax.yaxis.set_ticks_position('right')
# 	ax.spines['top'].set_position(('axes',0.25))
# 	ax.xaxis.set_ticks_position('top')
# 	ax.spines['left'].set_color('none')
# 	ax.spines['bottom'].set_color('none')
#




import tables as tb

def get_or_create_group(h5file, where, name=None, title='', filters=None, createparents=False, skip_on_readonly=False):
	if "/" in name:
		names = name.split('/')
		where += '/' + '/'.join(names[:-1])
		name = names[-1]
	try:
		return h5file.get_node(where, name=name)
	except tb.NoSuchNodeError:
		if name is not None:
			try:
				return h5file.create_group(where, name, title=title, filters=filters, createparents=createparents)
			except tb.FileModeError:
				if skip_on_readonly:
					return
		else:
			raise



def get_or_create_subgroup(parentgroup, name, title='', filters=None, skip_on_readonly=False):
	from .h5pod.generic import H5Pod
	if isinstance(parentgroup, H5Pod):
		parentgroup = parentgroup._groupnode
	if parentgroup is None and skip_on_readonly:
		return None
	h5file = parentgroup._v_file
	try:
		return h5file.get_node(parentgroup, name=name)
	except tb.NoSuchNodeError:
		if name is not None:
			try:
				return h5file.create_group(parentgroup, name, title=title, filters=filters, createparents=False)
			except tb.FileModeError:
				if skip_on_readonly:
					return
		else:
			raise






import logging
import numpy
import pandas
import os
from appdirs import user_cache_dir
import joblib
from typing import MutableMapping

from .general_precision import l4_float_dtype
from .log import logger_name
from .dataframes import DataFrames

def user_cache_file(filename, appname=None, appauthor=None, version=None, opinion=True):
	d = user_cache_dir(appname=appname, appauthor=appauthor, version=version, opinion=opinion)
	os.makedirs(d, exist_ok=True)
	return os.path.join(d, filename)


class Prelearner():
	"""
	A prelearner for use with Larch.

	A prelearner uses a machine learning classifier to make an initial
	prediction of the result.  This initial prediction is then added
	as an input data column for Larch, effectively creating a chained
	classifier.

	Parameters
	----------
	training_X : pandas.DataFrame
		The exogenous variables.
	training_Y : pandas.DataFrame
		The observed choices in the training data.
	training_W : pandas.DataFrame, optional
		The weights.
	classifier : sklearn Classifier or Regressor
		This is the class object for the selected classifier, not
		an existing instance.  This classifier or Regressor will be
		instantiated and trained using the data above to generate
		the prediction.
	fit : dict or False, optional
		A dictionary of arguments to pass to the `fit` method of the
		classifier, or set to False to not fit the classifier
		during the initialization of this object.
	cache_file : str, optional
		A cache file name to store the trained prelearner.  If just a filename is given,
		it will be stored in `appdirs.user_cache_file()`. If instead an absolute path or
		a relative path beginning with '.' is given, that location will be used.
		If the file exists, it will be loaded instead of re-training.
	output_name : str, default 'prelearned_utility'
		The name of the output column from this prelearner.
	grid_cv_params : dict or List[dict], optional
		If given, this is used as the `param_grid` argument
		to initialize a :class:`sklearn.model_selection.GridSearchCV`
		wrapped around the classifier, instead of using the
		classifier directly.
	grid_cv_kwds : dict, optional
		If `grid_cv_params` is given, this dict gives other keyword
		arguments given to :class:`sklearn.model_selection.GridSearchCV`.
	**kwargs
		Any other keyword arguments are passed through to the classifier's
		constructor.
	"""

	def __init__(
			self,
			dataframes,
			ca_columns=None,
			co_columns=None,
			classifier=None,
			fit=True,
			cache_file=None,
			output_name='prelearned_utility',
			appname='larch',
			grid_cv_params=None,
			grid_cv_kwds=None,
			validation_dataframes=None,
			**kwargs,
	):

		if classifier is None:
			raise ValueError('must give a classifier')

		if fit is True:
			fit = {}

		logger = logging.getLogger(logger_name)

		self.input_ca_columns = ca_columns if ca_columns is not None else []
		self.input_co_columns = co_columns

		self.eval_set_names = fit.pop('eval_set_names', [])
		
		if isinstance(fit, MutableMapping):
			if 'validation_percent' in fit and validation_dataframes is None:
				vpct = fit.pop('validation_percent')
				dataframes, validation_dataframes = dataframes.split([100-vpct, vpct])
			else:
				vpct = 'preset'
			if validation_dataframes is not None:
				validation_X = self.filter_and_join_columns(
					validation_dataframes.data_ca_as_ce(),
					validation_dataframes.data_co,
				)
				validation_Y = validation_dataframes.array_ch_as_ce()
				validation_W = validation_dataframes.array_wt_as_ce()

				fit['eval_set'] = fit.get('eval_set', []) + [(validation_X, validation_Y)]
				if validation_W is not None:
					fit['sample_weight_eval_set'] = fit.get('sample_weight_eval_set', []) + [validation_W]
				self.eval_set_names += [f'validation_{vpct}']

		training_X = self.filter_and_join_columns(
			dataframes.data_ca_as_ce(),
			dataframes.data_co,
		)
		training_Y = dataframes.array_ch_as_ce()
		training_W = dataframes.array_wt_as_ce()
		self.output_column = output_name

		if cache_file is not None:
			if os.path.isabs(cache_file) or cache_file[:2] in ('./', '..', '.\\'):
				cache_clf_file = cache_file
			else:
				cache_clf_file = user_cache_file(cache_file, appname=appname)
		else:
			cache_clf_file = None

		if cache_clf_file is not None and os.path.exists(cache_clf_file):
			logger.info(f'LOADING {cache_clf_file}...')
			clf = joblib.load(cache_clf_file)
			logger.info(f'COMPLETED LOADING {cache_clf_file}')
		else:
			if grid_cv_params is not None:
				from sklearn.model_selection import GridSearchCV
				clf = GridSearchCV(
					classifier(**kwargs),
					grid_cv_params,
					**grid_cv_kwds,
				)
			else:
				clf = classifier(**kwargs)
			if fit is not False:

				if 'train_as_eval' in fit:
					fit.pop('train_as_eval')
					fit['eval_set'] = [(training_X, training_Y),] + fit.get('eval_set',[])
					if training_W is not None:
						fit['sample_weight_eval_set'] = [training_W,]+fit.get('sample_weight_eval_set',[])
					self.eval_set_names = ['training'] + self.eval_set_names

				logger.info(f'FITTING {classifier}...')
				if training_W is not None:
					clf.fit(training_X, training_Y, sample_weight=training_W, **fit)
				else:
					clf.fit(training_X, training_Y, **fit)
				logger.info(f'FITTED {classifier}')
			if cache_clf_file is not None:
				joblib.dump(clf, cache_clf_file)
				logger.info(f'SAVED {cache_clf_file}')

		self.clf = clf
		self._predict_type = 'predict_proba col 1'


	def filter_ca_columns(self, X):
		# filter the columns of the input into the correct form for the prelearner.
		try:
			X1 = X[self.input_ca_columns]
		except KeyError:
			X1 = pandas.DataFrame(
				X.eval(self.input_ca_columns).T.astype(float),
				index=X.index,
				columns=self.input_ca_columns,
			)
		return X1

	def filter_and_join_columns(self, X_ca, X_co):
		training_X = self.filter_ca_columns(X_ca)
		if self.input_co_columns:
			try:
				X_co = X_co[self.input_co_columns]
			except KeyError:
				X_co = pandas.DataFrame(
					X_co.eval(self.input_co_columns).T.astype(float),
					index=X_co.index,
					columns=self.input_co_columns,
				)
			training_X = training_X.join(X_co, on=training_X.index.levels[0].name, how='left').fillna(0)
		return training_X

	def apply(
			self,
			X,
			dtype=None,
			output_name=None,
			**kwargs,
	):
		"""
		Apply the prelearner to compute pseudo-utility.

		Parameters
		----------
		X : pandas.DataFrame
		dtype : dtype, default float
			The dtype to use for the output column.
		output_name : str, optional
			The name of the output column from this
			application of the prelearner.
		**kwargs
			Other keyword arguments are forwarded to the
			`predict` or `predict_proba` method of the
			`clf` member.

		Returns
		-------
		pandas.DataFrame
		"""
		if dtype is None:
			dtype = l4_float_dtype

		if isinstance(X, DataFrames):
			X_ca = X._data_ca_or_ce
			X_co = X.data_co
		else:
			X_ca = X
			X_co = None

		X_in = self.filter_and_join_columns(
			X_ca,
			X_co,
		)

		if output_name is None:
			output_name = self.output_column

		if self._predict_type == 'predict_proba col 1':
			X_ca.loc[:,output_name] = numpy.log(self.clf.predict_proba(X_in, **kwargs)[:, 1]).astype(dtype)
		elif self._predict_type == 'predict':
			X_ca.loc[:,output_name] = numpy.log(self.clf.predict(X_in, **kwargs)).astype(dtype)
		else:
			raise TypeError(self._predict_type)
		return X



class RandomForestPrelearner(Prelearner):
	def __init__(
			self,
			dataframes,
			ca_columns=None,
			co_columns=None,
			cache_file=None,
			fit=True,
			output_name='prelearned_utility',
			**kwargs,
	):
		from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier

		default_kwargs = dict(
			n_estimators=200,
			warm_start=False,
			max_features=None,
			oob_score=True,
			n_jobs=-1,
			random_state=123,
		)

		default_kwargs.update(kwargs)

		super().__init__(
			dataframes=dataframes,
			ca_columns=ca_columns,
			co_columns=co_columns,
			classifier=RandomForestClassifier,
			fit=fit,
			cache_file=cache_file,
			output_name=output_name,
			**default_kwargs,
		)


class XGBoostHardPrelearner(Prelearner):

	def __init__(
			self,
			dataframes,
			ca_columns=None,
			co_columns=None,
			cache_file=None,
			fit=True,
			output_name='prelearned_utility',
			**kwargs,
	):
		from xgboost import XGBClassifier

		default_kwargs = dict(
			max_depth=11,
			learning_rate=0.01,
			n_estimators=500,
			silent=True,
			objective='binary:logistic',
			booster='gbtree',
			n_jobs=-1,
			max_delta_step=0,
			subsample=1,
			colsample_bytree=1,
			colsample_bylevel=1,
			reg_alpha=0,
			reg_lambda=1,
			scale_pos_weight=1,
			base_score=0.5,
			random_state=123,
		)

		default_kwargs.update(kwargs)

		super().__init__(
			dataframes=dataframes,
			ca_columns=ca_columns,
			co_columns=co_columns,
			classifier=XGBClassifier,
			fit=fit,
			cache_file=cache_file,
			output_name=output_name,
			**default_kwargs,
		)


class XGBoostSoftPrelearner(Prelearner):

	def __init__(
			self,
			dataframes,
			ca_columns=None,
			co_columns=None,
			cache_file=None,
			fit=True,
			output_name='prelearned_utility',
			**kwargs,
	):
		from xgboost import XGBRegressor

		default_kwargs = dict(
			max_depth=11,
			learning_rate=0.01,
			n_estimators=500,
			silent=True,
			objective='reg:logistic',
			booster='gbtree',
			n_jobs=-1,
			max_delta_step=0,
			subsample=1,
			colsample_bytree=1,
			colsample_bylevel=1,
			reg_alpha=0,
			reg_lambda=1,
			scale_pos_weight=1,
			base_score=0.5,
			random_state=123,
		)

		default_kwargs.update(kwargs)

		super().__init__(
			dataframes=dataframes,
			ca_columns=ca_columns,
			co_columns=co_columns,
			classifier=XGBRegressor,
			fit=fit,
			cache_file=cache_file,
			output_name=output_name,
			**default_kwargs,
		)

		self._predict_type = 'predict'




class XGBoostPrelearner(Prelearner):

	def __init__(
			self,
			dataframes,
			ca_columns=None,
			co_columns=None,
			cache_file=None,
			fit=True,
			output_name='prelearned_utility',
			**kwargs,
	):
		from xgboost import XGBRegressor, XGBClassifier

		training_Y = dataframes.array_ch_as_ce()
		use_soft = numpy.any((training_Y != 0) & (training_Y != 1.0))

		default_kwargs = dict(
			max_depth=11,
			learning_rate=0.01,
			n_estimators=500,
			silent=True,
			objective='reg:logistic' if use_soft else 'binary:logistic',
			booster='gbtree',
			n_jobs=-1,
			max_delta_step=0,
			subsample=1,
			colsample_bytree=1,
			colsample_bylevel=1,
			reg_alpha=0,
			reg_lambda=1,
			scale_pos_weight=1,
			base_score=0.5,
			random_state=123,
		)

		default_kwargs.update(kwargs)

		super().__init__(
			dataframes=dataframes,
			ca_columns=ca_columns,
			co_columns=co_columns,
			classifier=XGBRegressor if use_soft else XGBClassifier,
			fit=fit,
			cache_file=cache_file,
			output_name=output_name,
			**default_kwargs,
		)

		self._predict_type = 'predict' if use_soft else 'predict_proba col 1'


	def evals_result(self):
		j = [
			pandas.DataFrame({mk:numpy.asarray(mv) for mk, mv in ev.items()}) 
			for ek, ev in self.clf.evals_result_.items()
		]
		k = [
			ek
			for ek, ev in self.clf.evals_result_.items()
		]
		for i in range(len(self.eval_set_names)):
			if len(k)>i:
				k[i] = self.eval_set_names[i]
		return pandas.concat(j, axis=1, keys=k, sort=False)


import numpy
from ...util.aster import asterize
from ..general import _sqz_same, selector_len_for

class H5PodGroup():
	"""
	A DataPodGroup is a list of DataPods that all provide identically shaped data arrays.
	"""

	def __init__(self, members=None):
		self._pods = []
		if members is not None:
			for m in members:
				self.append(m)

	def append(self, x):
		from . import H5Pod
		if not isinstance(x, H5Pod):
			raise TypeError
		return self._pods.append(x)

	@property
	def shape(self):
		if len(self._pods):
			return self._pods[0].shape
		raise ValueError('no pods')

	def _getgroup(self, i):
		return self._pods[i]._groupnode

	def __getattr__(self, item):
		if item[:10]=='_groupnode' and len(item)>10:
			i = int(item[10:])
			return self._pods[i]._groupnode
		if item[:5]=='_pods' and len(item)>5:
			i = int(item[5:])
			return self._pods[i]
		raise AttributeError(item)

	def __len__(self):
		return len(self._pods)

	def _remake_command(self, cmd, selector=None, receiver=None):
		from tokenize import tokenize, untokenize, NAME, OP, STRING
		DOT = (OP, '.')
		COLON = (OP, ':')
		COMMA = (OP, ',')
		OBRAC = (OP, '[')
		CBRAC = (OP, ']')
		OPAR = (OP, '(')
		CPAR = (OP, ')')
		from io import BytesIO
		recommand = []

		if receiver:
			recommand += [(NAME, receiver), OBRAC, COLON, CBRAC, (OP, '='), ]

		try:
			cmd_encode = cmd.encode('utf-8')
		except AttributeError:
			cmd_encode = str(cmd).encode('utf-8')
		dims = len(self.shape)
		g = tokenize(BytesIO(cmd_encode).readline)
		if selector is None:
			screen_tokens = [COLON,]
		else:
			# try:
			# 	slicer_encode = selector.encode('utf-8')
			# except AttributeError:
			# 	slicer_encode = str(selector).encode('utf-8')
			# screen_tokens = [(toknum, tokval) for toknum, tokval, _, _, _ in tokenize(BytesIO(slicer_encode).readline)]
			screen_tokens = [(NAME, 'selector'), ]
		for toknum, tokval, _, _, _ in g:
			tok_accept = False
			for j in range(len(self)):
				if toknum == NAME and tokval in self._pods[j]:
					# replace NAME tokens
					partial = [(NAME, 'self'), DOT, (NAME, f'_pods{j}'), OBRAC, (STRING, f"'{tokval}'"), COMMA, ]
					partial += screen_tokens
					partial += [CBRAC, DOT, (NAME, f'squeeze'), OPAR, CPAR ]
					recommand.extend(partial)
					tok_accept = True
					break
			if not tok_accept:
				recommand.append((toknum, tokval))
		ret = untokenize(recommand).decode('utf-8')
		return asterize(ret, mode="exec" if receiver is not None else "eval"), ret


	def _evaluate_single_item(self, cmd, selector=None, receiver=None):
		j, j_raw = self._remake_command(cmd, selector=selector, receiver='receiver' if receiver is not None else None)
		# important globals
		from ...util.aster import inXd
		from numpy import log, exp, log1p, absolute, fabs, sqrt, isnan, isfinite, logaddexp, fmin, fmax, nan_to_num, sin, cos, pi
		from ...util.common_functions import piece, normalize, boolean
		try:
			if receiver is not None:
				exec(j)
			else:
				return eval(j)
		except Exception as exc:
			args = exc.args
			if not args:
				arg0 = ''
			else:
				arg0 = args[0]
			arg0 = arg0 + '\nwithin parsed command: "{!s}"'.format(cmd)
			if "max" in cmd:
				arg0 = arg0 + '\n(note to get the maximum of arrays use "fmax" not "max")'.format(cmd)
			if "min" in cmd:
				arg0 = arg0 + '\n(note to get the minimum of arrays use "fmin" not "min")'.format(cmd)
			if isinstance(exc, NameError):
				badname = str(exc).split("'")[1]
				goodnames = set(dir())
				from ...util.text_manip import case_insensitive_close_matches
				did_you_mean_list = case_insensitive_close_matches(badname, goodnames, n=3, cutoff=0.1, excpt=None)
				if len(did_you_mean_list) > 0:
					arg0 = arg0 + '\n' + "did you mean {}?".format(
						" or ".join("'{}'".format(s) for s in did_you_mean_list))
			exc.args = (arg0,) + args[1:]
			raise

	def __getitem__(self, item):
		if isinstance(item, tuple) and len(item)>=2 and isinstance(item[-1], slice):
			names, slice_ = item[:-1], item[-1]
		else:
			names = item
			slice_ = None

		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names,]

		dtype = numpy.float64

		result = numpy.zeros( [selector_len_for(slice_, self.shape[0]), *self.shape[1:], len(names)], dtype=dtype)

		for i, cmd in enumerate(names):
			result[...,i] = self._evaluate_single_item(cmd, slice_)

		return result

	def _load_into(self, names, slc, result=None, dtype=numpy.float64):
		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names, ]
		if result is None:
			# initialize a new array of the indicated dtype
			result = numpy.zeros([selector_len_for(slc, self.shape[0]), *self.shape[1:], len(names)], dtype=dtype)
		else:
			_sqz_same (result.shape,[selector_len_for(slc, self.shape[0]), *self.shape[1:], len(names)])
		for i, cmd in enumerate(names):
			result[..., i] = self._evaluate_single_item(cmd, slc)
		return result

	def load_into(self, names, selector=None, result_array=None, dtype=numpy.float64, mask=None):
		"""
		Load some data.

		Parameters
		----------
		names : iterable of str
			The variable names to load.
		selector : slice, optional
			A slice or bool array used to identify which rows to load.
		result_array : ndarray, optional
			The array to load the data into.  Must be the correct size. If not
			given, an array will be created.
		dtype : dtype, optional
			The dtype for the result_array to be created.
		mask : list, optional
			A list of the same shape as `names`.  For each item, if the mask evaluates
			as False, then the data will be skipped.  This is an optimization tool
			for reloading data that may not have changed (e.g. for logsum generation).

		Returns
		-------
		ndarray
		"""
		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names, ]
		if result_array is None:
			# initialize a new array of the indicated dtype
			result_array = numpy.zeros([selector_len_for(selector, self.shape[0]), *self.shape[1:], len(names)], dtype=dtype)
		else:
			_sqz_same(result_array.shape,[selector_len_for(selector, self.shape[0]), *self.shape[1:], len(names)])
		if mask is None:
			for i, cmd in enumerate(names):
				self._evaluate_single_item(cmd, selector, receiver=result_array[..., i])
		else:
			for i, cmd in enumerate(names):
				if mask[i]:
					self._evaluate_single_item(cmd, selector, receiver=result_array[..., i])
		return result_array

	def __repr__(self):
		s = f"<larch.{self.__class__.__name__}>"
		return s



from larch.model import Model
import numpy
from pytest import approx
import pandas
from larch.data_services.examples import ITINERARY_RAW
from larch.data_services import H5PodCE
from larch import DataService, Model
from larch.util.common_functions import fourier_series, piecewise_linear

import pytest

@pytest.mark.skip(reason="NNNL is not functioning correctly presently")
def test_2_tier():
	df = ITINERARY_RAW()

	df['outDepTimeHours'] = numpy.round(df['outDepTime'] / 3600).astype(int)

	pod_ce = H5PodCE.from_dataframe(
		df,
		caselabels='individual',
		altlabels='alternative',
	)

	_, sa = pod_ce.new_alternative_codes(
		groupby=[
			# 'nFlights',
			'outDepTimeHours',
		],
		name='AltsByFlightsDepHours',
		create_index='AltsByFlightsDepHours_idx',
		groupby_prefixes=[
			# "F",
			"H",
		],
		overwrite=True,
	)

	dx = DataService.from_CE_and_SA(pod_ce, sa)


	m = Model(dataservice=dx)

	from .. import P,X,PX

	m.utility_ca = (
			+ sum(PX(f'normalize({i})') for i in [
				'totalTripDurationMinutes',
				'log(totalPrice)',
				'nFlights',
			])
			+ fourier_series("outDepTime/86400", length=4)
			+ fourier_series("outArrTime/86400", length=4)
	)
	m.availability_var = 'is_avail'
	m.choice_ca_var = 'choice'
	m.magic_nesting(sa)
	m.unmangle()
	m.estimate()

	assert m.loglike() == approx(-1552.7201587407562)

	from larch.model.nnnl import NNNL
	n = NNNL(m)
	n.finalize()
	n.load_data()

	n.set_values(m.pf.value / m.pf.value['MU_outDepTimeHours'])
	n.set_value('MU_outDepTimeHours', m.pf.value['MU_outDepTimeHours'])

	assert n.loglike(n.pf.value) == approx( m.loglike(m.pf.value) )

	numpy.testing.assert_array_almost_equal(
		n.d_logsums(),
		m.d_logsums()
	)




@pytest.mark.skip(reason="NNNL is not functioning correctly presently")
def test_3_tier():
	df = ITINERARY_RAW()

	df['outDepTimeHours'] = numpy.round(df['outDepTime'] / 3600).astype(int)

	pod_ce = H5PodCE.from_dataframe(
		df,
		caselabels='individual',
		altlabels='alternative',
	)

	_, sa = pod_ce.new_alternative_codes(
		groupby=[
			'nFlights',
			'outDepTimeHours',
		],
		name='AltsByFlightsDepHours',
		create_index='AltsByFlightsDepHours_idx',
		groupby_prefixes=[
			"F",
			"H",
		],
		overwrite=True,
	)

	dx = DataService.from_CE_and_SA(pod_ce, sa)


	m = Model(dataservice=dx)

	from .. import P,X,PX

	m.utility_ca = (
			+ sum(PX(f'normalize({i})') for i in [
				'totalTripDurationMinutes',
				'log(totalPrice)',
				'nFlights',
			])
			+ fourier_series("outDepTime/86400", length=4)
			+ fourier_series("outArrTime/86400", length=4)
	)
	m.availability_var = 'is_avail'
	m.choice_ca_var = 'choice'
	m.magic_nesting(sa)
	m.unmangle()

	m.set_values({
		'MU_nFlights': 0.75,
		'MU_outDepTimeHours': 0.5,
		'cos_2πoutArrTime/86400': -0.24352756182332694,
		'cos_2πoutDepTime/86400': -0.5941495103991815,
		'cos_4πoutArrTime/86400': 0.07133782207978602,
		'cos_4πoutDepTime/86400': -0.45117462119982826,
		'normalize(log(totalPrice))': -1.225219579598701,
		'normalize(nFlights)': -2.3288120787774154,
		'normalize(totalTripDurationMinutes)': -0.44072158058922295,
		'sin_2πoutArrTime/86400': -1.0494189248239782,
		'sin_2πoutDepTime/86400': 0.06320548375577714,
		'sin_4πoutArrTime/86400': -0.4007737199632005,
		'sin_4πoutDepTime/86400': -0.15198982460097238,
	})

	m.load_data()

	assert approx(-1555.1201442643503) ==  m.loglike()

	from larch.model.nnnl import NNNL
	n = NNNL(m)
	n.finalize()
	n.load_data()

	n.set_values(m.pf.value / m.pf.value['MU_outDepTimeHours'])
	n.set_value('MU_outDepTimeHours', m.pf.value['MU_outDepTimeHours'] / m.pf.value['MU_nFlights'])
	n.set_value('MU_nFlights', m.pf.value['MU_nFlights'])

	assert n.loglike(n.pf.value) == approx( m.loglike(m.pf.value) )

	numpy.testing.assert_array_almost_equal(
		n.d_logsums(),
		m.d_logsums()
	)

